@@ -31,8 +31,12 @@ typedef int (*Fixup_Proc)(void *obj);
 typedef int (*Fixup2_Proc)(void *obj, struct NewGC *);
 typedef void (*GC_collect_start_callback_Proc)(void);
 typedef void (*GC_collect_end_callback_Proc)(void);
-typedef void (*GC_collect_inform_callback_Proc)(int major_gc, intptr_t pre_used, intptr_t post_used);
+typedef void (*GC_collect_inform_callback_Proc)(int master_gc, int major_gc, 
+                                                intptr_t pre_used, intptr_t post_used,
+                                                intptr_t pre_admin, intptr_t post_admin,
+                                                intptr_t post_child_places_used);
 typedef uintptr_t (*GC_get_thread_stack_base_Proc)(void);
+typedef void (*GC_Post_Propagate_Hook_Proc)(struct NewGC *);
 /* 
    Types of the traversal procs (supplied by Racket); see overview in README
    for information about traversals. The return value is the size of
@@ -87,7 +91,8 @@ GC2_EXTERN void GC_add_roots(void *start, void *end);
    `start' (inclusive) and `end' (exclusive) contains pointers. */
 
 GC2_EXTERN void GC_init_type_tags(int count, int pair, int mutable_pair, int weakbox, 
-                                  int ephemeron, int weakarray, int custbox);
+                                  int ephemeron, int weakarray, int custbox,
+                                  int phantom);
 /*
    Called by Racket to indicate the number of different type tags it
    uses, starting from 0. `count' is always less than 256. The weakbox
@@ -102,15 +107,18 @@ GC2_EXTERN void GC_register_root_custodian(void *);
 
 GC2_EXTERN void GC_register_new_thread(void *, void *);
 /*
-   Indicates that a just-allocated point is for a thread record
-   owned by a particular custodian. */
+   Indicates that a just-allocated point is for a thread 
+   or place owned by a particular custodian. */
+
 GC2_EXTERN void GC_register_thread(void *, void *);
 /*
-   Indicates that a a thread record is owned by a particular custodian. */
+   Indicates that a a thread or place is now owned by a
+   particular custodian. */
 
 GC2_EXTERN GC_collect_start_callback_Proc GC_set_collect_start_callback(GC_collect_start_callback_Proc);
 GC2_EXTERN GC_collect_end_callback_Proc GC_set_collect_end_callback(GC_collect_end_callback_Proc);
 GC2_EXTERN void GC_set_collect_inform_callback(GC_collect_inform_callback_Proc);
+GC2_EXTERN void GC_set_post_propagate_hook(GC_Post_Propagate_Hook_Proc);
 /*
    Sets callbacks called by GC before/after performing a collection.  Used by
    Racket to zero out some data and record collection times. The end
@@ -135,6 +143,10 @@ GC2_EXTERN intptr_t GC_get_memory_use(void *c);
    Returns the number of currently-allocated bytes (speficilly for
    custodian c, as much as the GC's accounting makes possible). */
 
+GC2_EXTERN int GC_accouting_enabled();
+/* 
+   Reports whether memory accounting is enabled. */
+
 #define MZACCT_REQUIRE		0
 #define MZACCT_LIMIT		1
 GC2_EXTERN int GC_set_account_hook(int type, void *c1, uintptr_t b, void *c2);
@@ -142,7 +154,17 @@ GC2_EXTERN int GC_set_account_hook(int type, void *c1, uintptr_t b, void *c2);
   Set a memory-accounting property. Returns 0 for failure (i.e., not
   supported). */
 
+GC2_EXTERN uintptr_t GC_get_account_memory_limit(void *c1);
+/*
+  Returns a moemory accounting limit for c1 (or any ancestor),
+  or 0 if none is set. */
+
 GC2_EXTERN void GC_gcollect(void);
+GC2_EXTERN void GC_gcollect_minor(void);
+/*
+   Performs an immediate [full] collection. */
+
+GC2_EXTERN void GC_enable_collection(int on);
 /*
    Performs an immediate (full) collection. */
 
@@ -182,19 +204,6 @@ GC2_EXTERN void *GC_malloc_pair(void *car, void *cdr);
    The main potential advantage is that `car' and `cdr' don't
    have to be retained by the callee in the case of a GC. */
 
-GC2_EXTERN void *GC_malloc_one_xtagged(size_t);
-/* 
-   Alloc an item, initially zeroed. Rather than having a specific tag,
-   all objects allocated this way are marked/fixedup via the function
-   in GC_mark_xtagged and GC_fixup_xtagged. Racket sets
-   GC_{mark,fixup}_xtagged. */
-
-GC2_EXTERN void (*GC_mark_xtagged)(void *obj);
-GC2_EXTERN void (*GC_fixup_xtagged)(void *obj);
-/* 
-  Mark and fixup functions for memory allocated with
-  GC_malloc_one_xtagged(). */
-
 GC2_EXTERN void *GC_malloc_array_tagged(size_t);
 /* 
    Alloc an array of tagged items. Racket sets the tag in the first
@@ -242,7 +251,7 @@ GC2_EXTERN void GC_free(void *);
    Lets the collector optionally reverse an allocation immediately.
    [Generally a no-op.] */
 
-  GC2_EXTERN void *GC_malloc_weak_box(void *p, void **secondary, int soffset, int is_late);
+GC2_EXTERN void *GC_malloc_weak_box(void *p, void **secondary, int soffset, int is_late);
 /* 
    Allocate a weak box. See README for details. */
 
@@ -268,6 +277,11 @@ GC2_EXTERN int GC_is_on_allocated_page(void *p);
    the GC allocates objects (although p may or may not
    be a valid pointer to the start of an alloctaed object). */
 
+GC2_EXTERN int GC_allocate_phantom_bytes(intptr_t);
+/* 
+   Returns 0 if allocation should fail due to a memory limit, 
+   1 otherwise. */
+
 /***************************************************************************/
 /* Memory tracing                                                          */
 /***************************************************************************/
@@ -333,6 +347,7 @@ GC2_EXTERN void GC_register_traversers2(short tag, Size2_Proc size, Mark2_Proc m
    which must be in words. */
 
 GC2_EXTERN void *GC_resolve(void *p);
+GC2_EXTERN void *GC_resolve2(void *p, struct NewGC *gc);
 /*
    Can be called by a traversal proc to get the current address of a
    object that might have been moved already. This is necessary, for
@@ -359,6 +374,19 @@ GC2_EXTERN void GC_fixup2(void *p, struct NewGC *gc);
    These procedures and variables are internal to the current
    implementation, and are *not* part of the "official" interface. */
 
+GC2_EXTERN int GC_is_marked2(const void *p, struct NewGC *gc);
+/*
+   Reports whether p has been marked. */
+
+GC2_EXTERN int GC_is_partial(struct NewGC *gc);
+/* 
+   Reports whether the current GC is a non-full collection. */
+
+GC2_EXTERN void GC_retract_only_mark_stack_entry(void *pf, struct NewGC *gc);
+/*
+   Used for very special collaboration with GC. */
+
+
 GC2_EXTERN void GC_mark_variable_stack(void **var_stack,
 				       intptr_t delta,
 				       void *limit,
@@ -407,29 +435,49 @@ GC2_EXTERN void GC_write_barrier(void *p);
    Explicit write barrier to ensure that a write-barrier signal is not
    triggered by a memory write.
 */
+
 GC2_EXTERN void GC_switch_out_master_gc();
 /*
    Makes the current GC the master GC.
    Creates a new place specific GC and links it to the master GC.
 */
-GC2_EXTERN void GC_construct_child_gc();
+
+GC2_EXTERN struct NewGC *GC_get_current_instance();
 /*
-   Creates a new place specific GC and links to the master GC.
+   Returns a representation of the current GC.
 */
 
+GC2_EXTERN void GC_construct_child_gc(struct NewGC *parent_gc, intptr_t limit);
+/*
+   Creates a new place-specific GC that is a child for memory-accounting
+   purposes of the give parent GC. If `limit' is not 0, set the maximum
+   amount of memory the new GC is supposed to use.
+*/
+
+GC2_EXTERN intptr_t GC_propagate_hierarchy_memory_use();
+/* 
+   Notifies the parent GC (if any) of memory use by the current GC
+   and its children. The result is total memory use. */
+
 GC2_EXTERN void GC_destruct_child_gc();
 /*
-   Destroys a place specific GC once the place has finished.
+   Destroys a place-specific GC once the place has finished.
 */
 
 GC2_EXTERN void *GC_switch_to_master_gc();
 /*
-   Switches to the master GC
+   Switches to the master GC.
 */
 
 GC2_EXTERN void GC_switch_back_from_master(void *gc);
 /*
-   Switches to back to gc from the master GC
+   Switches to back to gc from the master GC.
+*/
+
+GC2_EXTERN int GC_is_using_master();
+/*
+   Reports whether the master GC is in use after a non-master GC
+   has been created.
 */
 
 GC2_EXTERN intptr_t GC_alloc_alignment();
@@ -481,7 +529,7 @@ GC2_EXTERN void *GC_finish_message_allocator();
 
 GC2_EXTERN void GC_adopt_message_allocator(void *msg_memory);
 /*
-   Adopts the message memory captures by the sending place into
+   Adopts the message memory captured by the sending place into
    the current receiving place's gc
 */
 
@@ -491,6 +539,35 @@ GC2_EXTERN intptr_t GC_is_place();
    Otherwise returns 0;
 */
 
+GC2_EXTERN intptr_t GC_message_objects_size(void *msg_memory);
+/*
+ Returns the total size of all objects allocated by the message allocator
+ */
+
+GC2_EXTERN intptr_t GC_message_allocator_size(void *msg_memory);
+/*
+ Returns the total size of all memory pages allocated by the message allocator
+ */
+
+GC2_EXTERN void GC_dispose_short_message_allocator(void *msg_memory);
+/*
+ Disposes of small message allocators that were copied by the receiving place
+ */
+
+GC2_EXTERN void GC_destroy_orphan_msg_memory(void *msg_memory);
+/*
+ Used to destroys a message allocators that is still in the place channel queue when
+ the place channels finalizer is called.
+ */
+
+GC2_EXTERN void GC_report_unsent_message_delta(intptr_t amt);
+/* 
+ Report message-in-flight size changes to the GC. This functionality is exposed,
+ rather than built into GC_finish_message_allocator(), GC_adpot_message_allocator(),
+ GC_dispose_short_message_allocator(), and GC_destroy_orphan_msg_memory(), so that
+ meesages to the master GC can be limited as long as the unsent message tracking
+ is within a factor of 2 or so.
+ */
 
 
 # ifdef __cplusplus
