@@ -370,8 +370,8 @@ int BTC_bi_chan_mark(void *p, struct NewGC *gc)
     /* Race conditions here on `mem_size', and likely double counting
        when the same async channels are accessible from paired bi
        channels --- but those approximations are ok for accounting. */
-    account_memory(gc, gc->current_mark_owner, bc->sendch->mem_size, 0);
-    account_memory(gc, gc->current_mark_owner, bc->recvch->mem_size, 0);
+    account_memory(gc, gc->current_mark_owner, bc->link->sendch->mem_size, 0);
+    account_memory(gc, gc->current_mark_owner, bc->link->recvch->mem_size, 0);
   }
   return gc->mark_table[btc_redirect_bi_chan](p, gc);
 }
@@ -447,6 +447,9 @@ static void BTC_do_accounting(NewGC *gc)
     last = cur;
     while(cur) {
       int owner = custodian_to_owner_set(gc, cur);
+      uintptr_t save_count = gc->phantom_count;
+
+      gc->phantom_count = 0;
 
       gc->current_mark_owner = owner;
       GCDEBUG((DEBUGOUTF,"MARKING THREADS OF OWNER %i (CUST %p)\n", owner, cur));
@@ -458,6 +461,11 @@ static void BTC_do_accounting(NewGC *gc)
 
       last = cur;
       box = cur->global_next; cur = box ? SCHEME_PTR1_VAL(box) : NULL;
+
+      owner_table = gc->owner_table;
+      owner_table[owner]->memory_use = add_no_overflow(owner_table[owner]->memory_use, 
+                                                       gcBYTES_TO_WORDS(gc->phantom_count));
+      gc->phantom_count = save_count;
     }
 
     /* walk backward folding totals int parent */
@@ -470,7 +478,8 @@ static void BTC_do_accounting(NewGC *gc)
         int powner = custodian_to_owner_set(gc, parent);
 
         owner_table = gc->owner_table;
-        owner_table[powner]->memory_use += owner_table[owner]->memory_use;
+        owner_table[powner]->memory_use = add_no_overflow(owner_table[powner]->memory_use,
+                                                          owner_table[owner]->memory_use);
         owner_table[powner]->master_memory_use += owner_table[owner]->master_memory_use;
       }
 
