@@ -31,19 +31,19 @@ void launchgdb() {
   }
 
   pid_t fpid = fork();
-  if (fpid == 0)                // child
+  if (fpid == 0)                /* child */
   {
     char _pidstr[20];
     char *pidstr = (char*) _pidstr;
     snprintf(pidstr, 20, "%i", getpid());
     execl("/usr/bin/bash", "rgdb", pidstr);
   }
-  else if (fpid < 0)            // failed to fork
+  else if (fpid < 0)            /* failed to fork */
   {
     printf("Failed to fork\n");
     exit(1);
   }
-  else                                   // parent
+  else                                   /* parent */
   {
     kill(getpid(), SIGSTOP);
   }
@@ -60,7 +60,7 @@ void launchgdb() {
 }
 #endif
 
-void fault_handler(int sn, struct siginfo *si, void *ctx)
+void fault_handler(int sn, siginfo_t *si, void *ctx)
 {
   void *p = si->si_addr;
   /* quick access to SIGSEGV info in GDB */
@@ -125,7 +125,11 @@ void fault_handler(int sn, struct siginfo *si, void *ctx)
 
 /* ========== FreeBSD/NetBSD/OpenBSD signal handler ========== */
 /*  As of 2007/06/29, this is a guess for NetBSD!  */
-#if defined(__FreeBSD__) || defined(__FreeBSD_kernel__) || defined(__NetBSD__) || defined(__OpenBSD__)
+#if defined(__FreeBSD__) \
+ || defined(__FreeBSD_kernel__) \
+ || defined(__NetBSD__) \
+ || defined(__OpenBSD__) \
+ || defined(__DragonFly__)
 # include <signal.h>
 # include <sys/param.h>
 void fault_handler(int sn, siginfo_t *si, void *ctx)
