@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2012 PLT Scheme Inc.
+  Copyright (c) 2004-2013 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
   All rights reserved.
 
@@ -63,6 +63,39 @@
 # define MZ_USE_SINGLE_FLOATS
 #endif
 
+/* gcc defines __SSE_MATH__ when SSE floating point is enabled: */
+#ifdef __SSE_MATH__
+# define C_COMPILER_USES_SSE 1
+#endif
+
+#ifdef C_COMPILER_USES_SSE
+# if defined(MZ_TRY_EXTFLONUMS) && !defined(MZ_NO_EXTFLONUMS)
+#  define MZ_LONG_DOUBLE
+#  ifdef ASM_DBLPREC_CONTROL_87
+#   define ASM_EXTPREC_CONTROL_87
+#  endif
+# endif
+# ifdef ASM_DBLPREC_CONTROL_87
+#  undef ASM_DBLPREC_CONTROL_87
+# endif
+# if defined(MZ_USE_JIT_I386) && !defined(MZ_NO_JIT_SSE)
+#  define MZ_USE_JIT_SSE
+# endif
+#endif
+
+#ifdef MZ_LONG_DOUBLE
+# if defined(_MSC_VER)
+#  define BYTES_RESERVED_FOR_LONG_DOUBLE 16
+typedef struct {
+  char bytes[BYTES_RESERVED_FOR_LONG_DOUBLE];
+} mz_long_double;
+# else
+typedef long double mz_long_double;
+# endif
+#else
+typedef double mz_long_double;
+#endif
+
 #ifdef DONT_ITIMER
 # undef USE_ITIMER
 #endif
@@ -110,6 +143,9 @@ typedef long FILE;
 #ifndef MSC_IZE
 # define MSC_IZE(x) x
 #endif
+#ifndef M_MSC_IZE
+# define M_MSC_IZE(x) x
+#endif
 #ifndef MSCBOR_IZE
 # define MSCBOR_IZE(x) MSC_IZE(x)
 #endif
@@ -127,10 +163,14 @@ typedef long FILE;
    old way. */
 #ifdef MZ_USE_OLD_ARRAY_STYLE
 # define mzFLEX_ARRAY_DECL 1
+# define mzFLEX_ARRAY4_DECL 4
 # define mzFLEX_DELTA 1
+# define mzFLEX4_DELTA 4
 #else
 # define mzFLEX_ARRAY_DECL /* empty */
+# define mzFLEX_ARRAY4_DECL /* empty */
 # define mzFLEX_DELTA 0
+# define mzFLEX4_DELTA 0
 #endif
 
 #ifdef MZ_XFORM
@@ -237,8 +277,8 @@ typedef _int64 mzlonglong;
 typedef unsigned _int64 umzlonglong;
 #else
 # if defined(NO_LONG_LONG_TYPE) || defined(SIXTY_FOUR_BIT_INTEGERS)
-typedef long mzlonglong;
-typedef unsigned long umzlonglong;
+typedef intptr_t mzlonglong;
+typedef uintptr_t umzlonglong;
 # else
 typedef long long mzlonglong;
 typedef unsigned long long umzlonglong;
@@ -321,6 +361,18 @@ typedef struct {
   double double_val;
 } Scheme_Double;
 
+#ifdef MZ_LONG_DOUBLE
+typedef struct {
+  Scheme_Object so;
+  mz_long_double long_double_val;
+} Scheme_Long_Double;
+#else
+typedef struct {
+  Scheme_Object so;
+  const char *printed_form;
+} Scheme_Long_Double;
+#endif 
+
 #ifdef MZ_USE_SINGLE_FLOATS
 typedef struct {
   Scheme_Object so;
@@ -331,7 +383,7 @@ typedef struct {
 typedef struct Scheme_Symbol {
   Scheme_Inclhash_Object iso; /* 1 in low bit of keyex indicates uninterned */
   intptr_t len;
-  char s[4]; /* Really, a number of chars to match `len' */
+  char s[mzFLEX_ARRAY4_DECL];
 } Scheme_Symbol;
 
 typedef struct Scheme_Vector {
@@ -350,6 +402,14 @@ typedef struct Scheme_Double_Vector {
   double els[mzFLEX_ARRAY_DECL];
 } Scheme_Double_Vector;
 
+#ifdef MZ_LONG_DOUBLE
+typedef struct Scheme_Long_Double_Vector {
+  Scheme_Inclhash_Object iso; /* & 0x2 indicates allocated in the MASTERGC */
+  intptr_t size;
+  mz_long_double els[mzFLEX_ARRAY_DECL];
+} Scheme_Long_Double_Vector;
+#endif
+
 typedef struct Scheme_Print_Params Scheme_Print_Params;
 typedef void (*Scheme_Type_Printer)(Scheme_Object *v, int for_display, Scheme_Print_Params *pp);
 
@@ -409,6 +469,8 @@ typedef intptr_t (*Scheme_Secondary_Hash_Proc)(Scheme_Object *obj, void *cycle_d
 #define SCHEME_REALP(obj)  (SCHEME_INTP(obj) || ((_SCHEME_TYPE(obj) >= scheme_bignum_type) && (_SCHEME_TYPE(obj) < scheme_complex_type)))
 #define SCHEME_NUMBERP(obj)  (SCHEME_INTP(obj) || ((_SCHEME_TYPE(obj) >= scheme_bignum_type) && (_SCHEME_TYPE(obj) <= scheme_complex_type)))
 
+#define SCHEME_LONG_DBLP(obj)     SAME_TYPE(SCHEME_TYPE(obj), scheme_long_double_type)
+
 #define SCHEME_CHAR_STRINGP(obj)  SAME_TYPE(SCHEME_TYPE(obj), scheme_char_string_type)
 #define SCHEME_MUTABLE_CHAR_STRINGP(obj)  (SCHEME_CHAR_STRINGP(obj) && SCHEME_MUTABLEP(obj))
 #define SCHEME_IMMUTABLE_CHAR_STRINGP(obj)  (SCHEME_CHAR_STRINGP(obj) && SCHEME_IMMUTABLEP(obj))
@@ -462,6 +524,7 @@ typedef intptr_t (*Scheme_Secondary_Hash_Proc)(Scheme_Object *obj, void *cycle_d
 #define SCHEME_IMMUTABLE_VECTORP(obj)  (SCHEME_VECTORP(obj) && SCHEME_IMMUTABLEP(obj))
 
 #define SCHEME_FLVECTORP(obj)  SAME_TYPE(SCHEME_TYPE(obj), scheme_flvector_type)
+#define SCHEME_EXTFLVECTORP(obj)  SAME_TYPE(SCHEME_TYPE(obj), scheme_extflvector_type)
 #define SCHEME_FXVECTORP(obj)  SAME_TYPE(SCHEME_TYPE(obj), scheme_fxvector_type)
 
 #define SCHEME_STRUCTP(obj) (SAME_TYPE(SCHEME_TYPE(obj), scheme_structure_type) || SAME_TYPE(SCHEME_TYPE(obj), scheme_proc_struct_type))
@@ -528,6 +591,9 @@ typedef intptr_t (*Scheme_Secondary_Hash_Proc)(Scheme_Object *obj, void *cycle_d
 #define SCHEME_CHAR_VAL(obj) (((Scheme_Small_Object *)(obj))->u.char_val)
 #define SCHEME_INT_VAL(obj)  (OBJ_TO_LONG(obj)>>1)
 #define SCHEME_DBL_VAL(obj)  (((Scheme_Double *)(obj))->double_val)
+#ifdef MZ_LONG_DOUBLE
+#define SCHEME_LONG_DBL_VAL(obj)  (((Scheme_Long_Double *)(obj))->long_double_val)
+#endif
 #ifdef MZ_USE_SINGLE_FLOATS
 # define SCHEME_FLT_VAL(obj)  (((Scheme_Float *)(obj))->float_val)
 # define SCHEME_FLOAT_VAL(obj) (SCHEME_DBLP(obj) ? SCHEME_DBL_VAL(obj) : SCHEME_FLT_VAL(obj))
@@ -574,6 +640,11 @@ typedef intptr_t (*Scheme_Secondary_Hash_Proc)(Scheme_Object *obj, void *cycle_d
 #define SCHEME_FLVEC_SIZE(obj) (((Scheme_Double_Vector *)(obj))->size)
 #define SCHEME_FLVEC_ELS(obj)  (((Scheme_Double_Vector *)(obj))->els)
 
+#ifdef MZ_LONG_DOUBLE
+#define SCHEME_EXTFLVEC_SIZE(obj) (((Scheme_Long_Double_Vector *)(obj))->size)
+#define SCHEME_EXTFLVEC_ELS(obj)  (((Scheme_Long_Double_Vector *)(obj))->els)
+#endif
+
 #define SCHEME_FXVEC_SIZE(obj) SCHEME_VEC_SIZE(obj)
 #define SCHEME_FXVEC_ELS(obj) SCHEME_VEC_ELS(obj)
 
@@ -660,16 +731,14 @@ typedef struct Scheme_Offset_Cptr
    Do not use them directly. */
 #define SCHEME_PRIM_OPT_MASK (1 | 2)
 #define SCHEME_PRIM_IS_PRIMITIVE 4
-#define SCHEME_PRIM_IS_UNSAFE_OMITABLE 8
-#define SCHEME_PRIM_IS_OMITABLE 16
+#define SCHEME_PRIM_IS_MULTI_RESULT 8
+#define SCHEME_PRIM_IS_CLOSURE 16
 #define SCHEME_PRIM_OTHER_TYPE_MASK (32 | 64 | 128 | 256)
-#define SCHEME_PRIM_IS_MULTI_RESULT 512
-#define SCHEME_PRIM_IS_BINARY_INLINED 1024
-#define SCHEME_PRIM_IS_UNSAFE_FUNCTIONAL 2048
-#define SCHEME_PRIM_IS_METHOD 4096
-#define SCHEME_PRIM_IS_CLOSURE 8192
-#define SCHEME_PRIM_IS_UNARY_INLINED 16384
-#define SCHEME_PRIM_IS_NARY_INLINED 32768
+#define SCHEME_PRIM_IS_METHOD 512
+
+#define SCHEME_PRIM_OPT_INDEX_SIZE 6
+#define SCHEME_PRIM_OPT_INDEX_SHIFT 10
+#define SCHEME_PRIM_OPT_INDEX_MASK ((1 << SCHEME_PRIM_OPT_INDEX_SIZE) - 1)
 
 /* Values with SCHEME_PRIM_OPT_MASK, earlier implies later: */
 #define SCHEME_PRIM_OPT_FOLDING    3
@@ -679,6 +748,7 @@ typedef struct Scheme_Offset_Cptr
 /* Values with SCHEME_PRIM_OTHER_TYPE_MASK */
 #define SCHEME_PRIM_STRUCT_TYPE_INDEXLESS_GETTER (32 | 256)
 #define SCHEME_PRIM_STRUCT_TYPE_CONSTR           128
+#define SCHEME_PRIM_STRUCT_TYPE_SIMPLE_CONSTR    (32 | 64 | 128)
 #define SCHEME_PRIM_STRUCT_TYPE_INDEXLESS_SETTER 256
 #define SCHEME_PRIM_STRUCT_TYPE_INDEXED_SETTER   (128 | 256)
 #define SCHEME_PRIM_STRUCT_TYPE_BROKEN_INDEXED_SETTER   (32 | 128)
@@ -1900,7 +1970,7 @@ MZ_EXTERN Scheme_Object *scheme_register_parameter(Scheme_Prim *function, char *
 /*                              addrinfo                                  */
 /*========================================================================*/
 
-#ifdef HAVE_GETADDRINFO
+#if defined(HAVE_GETADDRINFO) || defined(__MINGW32__)
 # define mz_addrinfo addrinfo
 #else
 struct mz_addrinfo {
