@@ -82,7 +82,6 @@
 # include "uconfig.h"
 
 # define USE_EXPLICT_FP_FORM_CHECK
-# define POW_HANDLES_INF_CORRECTLY
 
 # include <errno.h>
 # ifdef ECHRNG
@@ -144,8 +143,6 @@
 
 # define SELECT_INCLUDE
 
-# define POW_HANDLES_INF_CORRECTLY
-
 # define USE_FCNTL_O_NONBLOCK
 
 # define USE_TIMEZONE_VAR_W_DLS
@@ -162,9 +159,8 @@
 # if defined(i386)
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "i386-linux"
 #  define REGISTER_POOR_MACHINE
-#  ifndef MZ_USE_JIT_SSE
-#   define ASM_DBLPREC_CONTROL_87
-#  endif
+#  define MZ_TRY_EXTFLONUMS
+#  define ASM_DBLPREC_CONTROL_87
 # endif
 # if defined(powerpc)
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "ppc-linux"
@@ -188,9 +184,8 @@
 # if defined(__x86_64__)
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "x86_64-linux"
 #  define REGISTER_POOR_MACHINE
-#  ifdef MZ_NO_JIT_SSE
-#   define ASM_DBLPREC_CONTROL_87
-#  endif
+#  define ASM_DBLPREC_CONTROL_87
+#  define MZ_TRY_EXTFLONUMS
 # endif
 # ifndef SCHEME_PLATFORM_LIBRARY_SUBPATH
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "unknown-linux"
@@ -261,7 +256,8 @@
 #endif
 
 # define USE_IEEE_FP_PREDS
-# define POW_HANDLES_INF_CORRECTLY
+
+# define USE_DYNAMIC_FDSET_SIZE
 
 # define SIGSET_IS_SIGNAL
 
@@ -299,17 +295,28 @@
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "i386-openbsd"
 # endif
 
+# include <sys/param.h>
+# if OpenBSD < 201211
+/* This is needed for (pre-5.2) userspace threads: */
+#  define ASSUME_FIXED_STACK_SIZE
+#  define FIXED_STACK_SIZE 1048576
+# endif
+
 # include "uconfig.h"
 # undef HAS_STANDARD_IOB
-
 # define HAS_BSD_IOB
 
+/* Default UNIX_STACK_MAXIMUM is too big for a non-root user. */
+# undef UNIX_STACK_MAXIMUM
+# define UNIX_STACK_MAXIMUM 4194304
+
 #ifndef __ELF__
 # define UNDERSCORE_DYNLOAD_SYMBOL_PREFIX
 #endif
 
+# define USE_DLOPEN_GLOBAL_BY_DEFAULT
+
 # define USE_IEEE_FP_PREDS
-# define POW_HANDLES_INF_CORRECTLY
 
 # define USE_DYNAMIC_FDSET_SIZE
 
@@ -323,9 +330,13 @@
 #if defined(__x86_64__)
 # define MZ_USE_JIT_X86_64
 # define MZ_JIT_USE_MPROTECT
-#else
+# define MZ_TRY_EXTFLONUMS
+#elif defined(__i386__)
 # define MZ_USE_JIT_I386
 # define MZ_JIT_USE_MPROTECT
+# define MZ_TRY_EXTFLONUMS
+#else
+# error Unported platform.
 #endif
 
 # define FLAGS_ALREADY_SET
@@ -340,21 +351,19 @@
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "i386-freebsd"
 #  define REGISTER_POOR_MACHINE
 #  define MZ_USE_JIT_I386
-#  ifndef MZ_JIT_X86_SSE
-#    if defined(__FreeBSD_kernel__)
-#     define ASM_DBLPREC_CONTROL_87
-#    else
-#     define FREEBSD_CONTROL_387
-#    endif
+#  define MZ_TRY_EXTFLONUMS
+#  if defined(__FreeBSD_kernel__)
+#   define ASM_DBLPREC_CONTROL_87
+#  else
+#   define FREEBSD_CONTROL_387
 #  endif
 # elif defined(__amd64__)
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "amd64-freebsd"
 #  define REGISTER_POOR_MACHINE
 #  define MZ_USE_JIT_X86_64
-#  ifdef MZ_NO_JIT_SSE
-#    if defined(__FreeBSD_kernel__)
-#     define ASM_DBLPREC_CONTROL_87
-#    endif
+#  define MZ_TRY_EXTFLONUMS
+#  if defined(__FreeBSD_kernel__)
+#   define ASM_DBLPREC_CONTROL_87
 #  endif
 # elif defined(__sparc64__)
 #  define SCHEME_PLATFORM_LIBRARY_SUBPATH "sparc64-freebsd"
@@ -363,7 +372,7 @@
 #  error Unported platform.
 # endif
 
-/* pthreads always enabled via configure', and
+/* pthreads always enabled via `configure', and
    initial pthread's stack size doesn't use rlimit: */
 # define ASSUME_FIXED_STACK_SIZE
 # define FIXED_STACK_SIZE 1048576
@@ -380,9 +389,6 @@
 # define USE_UNDERSCORE_SETJMP
 
 # define USE_IEEE_FP_PREDS
-# ifndef ASM_DBLPREC_CONTROL_87
-#  define POW_HANDLES_INF_CORRECTLY
-# endif
 
 # define USE_DYNAMIC_FDSET_SIZE
 
@@ -518,7 +524,7 @@
   /* See the "worksp" directory for more MSVC details.       */
 
 #if (defined(__BORLANDC__) \
-     || (defined(_MSC_VER) \
+     || ((defined(_MSC_VER) || defined(__MINGW32__)) \
          && (defined(__WIN32__) || defined(WIN32) || defined(_WIN32))))
 
 # ifdef _WIN64
@@ -529,7 +535,7 @@
 
 # define SYSTEM_TYPE_NAME "windows"
 # define DOS_FILE_SYSTEM
-# if defined(_MSC_VER)
+# if defined(_MSC_VER) || defined(__MINGW32__)
 #  define NO_READDIR
 #  define USE_FINDFIRST
 #  define NO_READLINK
@@ -587,6 +593,10 @@
 # define TRIG_ZERO_NEEDS_SIGN_CHECK
 # define NEED_TO_DEFINE_MATHERR
 #endif
+#ifdef __MINGW32__
+# define USE_IEEE_FP_PREDS
+# define ATAN2_DOESNT_WORK_WITH_INFINITIES
+#endif
 
 # define IO_INCLUDE
 # define DONT_IGNORE_PIPE_SIGNAL
@@ -595,10 +605,13 @@
 # define NO_SLEEP
 # define WINDOWS_PROCESSES
 # define WINDOWS_FILE_HANDLES
-# define USE_WIN32_THREAD_TIMER
+# ifndef MZ_USE_DETERMINSTIC_FUEL
+#  define USE_WIN32_THREAD_TIMER
+# endif
 
 # define SIGSET_IS_SIGNAL
 # define SIGSET_NEEDS_REINSTALL
+# define NO_SIGHUP_HANDLER
 
 #define USE_WINSOCK_TCP
 
@@ -607,7 +620,14 @@
 /* MS Visual C++ likes underscore prefixes */
 #if defined(_MSC_VER)
 # define MSC_IZE(x) _ ## x
+# define M_MSC_IZE(x) MSC_IZE(x)
 # define DIRECT_INCLUDE
+# else
+# define M_MSC_IZE(x) x
+#endif
+
+#if defined(__MINGW32__)
+# define MSC_IZE(x) _ ## x
 #endif
 
 #ifdef __BORLANDC__
@@ -617,7 +637,13 @@
 #endif
 
 #if defined(_MSC_VER)
+# define MZ_LONG_DOUBLE
 # define IGNORE_BY_MS_CONTROL_87
+# define MZ_NEED_SET_EXTFL_MODE
+#endif
+#if defined(__MINGW32__)
+# define MZ_TRY_EXTFLONUMS
+# define ASM_DBLPREC_CONTROL_87
 #endif
 
 # define REGISTER_POOR_MACHINE
@@ -631,6 +657,7 @@
 # else
 #  define MZ_USE_JIT_I386
 # endif
+
 # define MZ_JIT_USE_WINDOWS_VIRTUAL_ALLOC
 
 # define FLAGS_ALREADY_SET
@@ -726,8 +753,10 @@
 # define SYSTEM_TYPE_NAME "macosx"
 #endif
 
-# undef USE_ITIMER
-# define USE_PTHREAD_THREAD_TIMER
+# ifndef MZ_USE_DETERMINSTIC_FUEL
+#  undef USE_ITIMER
+#  define USE_PTHREAD_THREAD_TIMER
+# endif
 
 # define USE_MAP_ANON
 
@@ -738,6 +767,8 @@
 
 # define SIGSET_IS_SIGNAL
 
+# define MZ_TCP_LISTEN_IPV6_ONLY_SOCKOPT
+
 # define USE_TM_GMTOFF_FIELD
 # define USE_TM_ZONE_FIELD
 
@@ -755,21 +786,19 @@
 
 #if defined(__POWERPC__)
 # define MZ_USE_JIT_PPC
-#elif defined(__x86_64__)
-# define MZ_USE_JIT_X86_64
 #else
-# define MZ_USE_JIT_I386
-# ifndef MZ_NO_JIT_SSE
-#  define MZ_USE_JIT_SSE
-#  define ASM_DBLPREC_CONTROL_87
+# if defined(__x86_64__)
+#  define MZ_USE_JIT_X86_64
+# else
+#  define MZ_USE_JIT_I386
 # endif
+# define ASM_DBLPREC_CONTROL_87
+# define MZ_TRY_EXTFLONUMS
 #endif
-# define MZ_JIT_USE_MPROTECT
 
-#if defined(__x86_64__)
-/* work around a bug in localtime() in 10.6.8 */
-# define MIN_VALID_DATE_SECONDS -67768040609715600
-#endif
+# define POW_HANDLES_CASES_CORRECTLY
+
+# define MZ_JIT_USE_MPROTECT
 
 # define FLAGS_ALREADY_SET
 
@@ -856,7 +885,6 @@
 # define NO_STAT_PROC
 # define DONT_IGNORE_PIPE_SIGNAL
 
-# define POW_HANDLES_INF_CORRECTLY
 # define TRIG_ZERO_NEEDS_SIGN_CHECK
 
 # define MACOS_UNICODE_SUPPORT
@@ -942,7 +970,6 @@
 # define DONT_IGNORE_PIPE_SIGNAL
 # define DONT_IGNORE_FPE_SIGNAL
 
-# define POW_HANDLES_INF_CORRECTLY
 # define USE_PALM_INF_TESTS
 
 # define FLAGS_ALREADY_SET
@@ -968,6 +995,10 @@
 
 # define FLAGS_ALREADY_SET
 
+# define USE_DYNAMIC_FDSET_SIZE
+
+# define BROKEN_READLINK_NUL_TERMINATOR
+
 #if defined(i386)
 # define MZ_USE_JIT_I386
 # define MZ_JIT_USE_MPROTECT
@@ -978,6 +1009,56 @@
 # define MZ_USE_DWARF_LIBUNWIND
 #endif
 
+#endif
+
+  /************ Dragonfly *************/
+
+#if defined(__DragonFly__)
+ 
+# if defined(__i386__)
+#  define SCHEME_PLATFORM_LIBRARY_SUBPATH "i386-dragonfly"
+#  define REGISTER_POOR_MACHINE
+#  define MZ_USE_JIT_I386
+#  define ASM_DBLPREC_CONTROL_87
+#  define MZ_TRY_EXTFLONUMS
+# elif defined(__amd64__)
+#  define SCHEME_PLATFORM_LIBRARY_SUBPATH "amd64-dragonfly"
+#  define REGISTER_POOR_MACHINE
+#  define MZ_USE_JIT_X86_64
+#  define ASM_DBLPREC_CONTROL_87
+#  define MZ_TRY_EXTFLONUMS
+# else
+#  error Unported platform.
+# endif
+
+/* pthreads always enabled via `configure', and
+   initial pthread's stack size doesn't use rlimit: */
+# define ASSUME_FIXED_STACK_SIZE
+# define FIXED_STACK_SIZE 1048576
+
+# include "uconfig.h"
+# undef HAS_STANDARD_IOB
+
+# define HAS_BSD_IOB
+
+# define USE_UNDERSCORE_SETJMP
+
+# define USE_IEEE_FP_PREDS
+# ifndef ASM_DBLPREC_CONTROL_87
+#  define POW_HANDLES_INF_CORRECTLY
+# endif
+
+# define USE_DYNAMIC_FDSET_SIZE
+
+# define SIGSET_IS_SIGNAL
+
+# define USE_TM_GMTOFF_FIELD
+# define USE_TM_ZONE_FIELD
+
+# define MZ_JIT_USE_MPROTECT
+
+# define FLAGS_ALREADY_SET
+
 #endif
 
   /***************************************************/
@@ -1079,6 +1160,10 @@
 
  /* NO_READLINK means that there is no readlink() function. */
 
+ /* BROKEN_READLINK_NUL_TERMINATOR means that readlink() may
+    report a length that includes trailing NUL terminators,
+    which should be stripped away. */
+
  /* USE_GETDISK uses getdisk() and setdisk() to implement the
      filesystem-root-list primitive under DOS. */
 
@@ -1277,10 +1362,32 @@
     converts (a == a) to TRUE, even if `a' is floating-point. Used
     only when USE_[SCO_]IEEE_FP_PREDS is not defined. */
 
+ /* C_COMPILER_USES_SSE indicates that the C compiler generates SSE
+    instructions for `double' arithmetic. This flag is turned on
+    automatically if __SSE_MATH__ is defined (usually by gcc). */
+
+ /* MZ_LONG_DOUBLE enables extflonum support. */
+
+ /* MZ_TRY_EXTFLONUMS turns on MZ_LONG_DOUBLE if C_COMPILER_USES_SSE. */
+
  /* ASM_DBLPREC_CONTROL_87 uses inline assembly to set Intel '387
+    floating-point operations to double-precision instead of
+    extended-precision arithmetic. This definition is turned off if
+    C_COMPILER_USES_SSE, and ASM_EXTPREC_CONTROL_87 is turned on
+    instead if C_COMPILER_USES_SSE and MZ_LONG_DOUBLE. */
+
+ /* ASM_EXTPREC_CONTROL_87 uses inline assembly to set Intel '387
     floating-point operations to double-precision instead of
     extended-precision arithmetic. */
 
+ /* MZ_NEED_SET_EXTFL_MODE causes JIT-generated extflonum instructions
+    to set the x87 control word to extended precision just before an
+    extflonum operation, and then set if back to double precision just
+    after. This is needed or Windows (where the default mode is double
+    precision, something about the x64 environment switches the mode
+    back if you try to change it permanently, and "longdouble.dll"
+    does the same switch for non-JITted operations). */
+
  /* IGNORE_BY_BORLAND_CONTROL_87 turns off floating-point error for
     Intel '387 with Borlad-style _control87. DONT_IGNORE_PIPE_SIGNAL
     can be on or off. */
@@ -1318,9 +1425,9 @@
     SunOS/Solaris, and HP/UX by explicit pre-checking the form of the 
     number and looking for values that are obviously +inf.0 or -inf.0 */
 
- /* POW_HANDLES_INF_CORRECTLY inidicates that thw pow() library procedure
-    handles +/-inf.0 correctly. Otherwise, code in inserted to specifically
-    check for infinite arguments. */
+ /* POW_HANDLES_CASES_CORRECTLY indicates that the pow() library procedure
+    handles all +/-inf.0, +/-0.0, or +nan.0 cases according to C99. This
+    might save time on redundant checking before Racket calls pow(). */
     
  /* ATAN2_DOESNT_WORK_WITH_INFINITIES indicates that atan2(+/-inf, +/-inf)
     is not the same as atan2(1, 1). */ 
@@ -1419,6 +1526,9 @@
     an extra underscore ("_") must be placed in front of the name passed 
     to dlopen(). */
 
+ /* USE_DLOPEN_GLOBAL_BY_DEFAULT opens shared libraries in "global"
+    mode by default, instead of "local" mode. */
+
  /* LINK_EXTENSIONS_BY_TABLE specifies that the Racket functions
     used by an extension must be manually linked via a table of
     function pointers. Windows dynamic linking uses this method. */
@@ -1503,7 +1613,12 @@
  /* NO_INLINE_KEYWORD indicates that the C compiler doesn't recognize
     C's `inline' keyword. */
 
- /* NO_USER_BREAK_HANDLER turns off handling of INT signal in main.c */
+ /* NO_USER_BREAK_HANDLER turns off handling of SIGINT, SIGTERM, and 
+    SIGHUP in main.c */
+
+ /* NO_SIGTERM_HANDLER turns off handling of SIGTERM in main.c */
+
+ /* NO_SIGHUP_HANDLER turns off handling of SIGHUP in main.c */
 
  /* DIR_INCLUDE if there's a <dir.h> file (mainly for Windows). */
 
