@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2012 PLT Scheme Inc.
+  Copyright (c) 2004-2013 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -101,6 +101,12 @@ void scheme_init_compenv_places(void)
   }
 }
 
+void scheme_init_compenv_symbol(void)
+{
+  REGISTER_SO(unshadowable_symbol);
+  unshadowable_symbol = scheme_intern_symbol("unshadowable");
+}
+
 /*========================================================================*/
 /*                       compilation info management                      */
 /*========================================================================*/
@@ -535,6 +541,11 @@ int scheme_is_toplevel(Scheme_Comp_Env *env)
   return !env->next || (env->flags & SCHEME_TOPLEVEL_FRAME);
 }
 
+int scheme_is_nested_module(Scheme_Comp_Env *env)
+{
+  return (env->flags & SCHEME_NESTED_MODULE_FRAME);
+}
+
 int scheme_is_module_env(Scheme_Comp_Env *env)
 {
   return !!(env->flags & SCHEME_MODULE_BEGIN_FRAME); /* name is backwards compared to symbol! */
@@ -850,16 +861,8 @@ Scheme_Object *scheme_make_local(Scheme_Type type, int pos, int flags)
   k = type - scheme_local_type;
   
   /* Helper for reading bytecode: make sure flags is a valid value */
-  switch (flags) {
-  case 0:
-  case SCHEME_LOCAL_CLEAR_ON_READ:
-  case SCHEME_LOCAL_OTHER_CLEARS:
-  case SCHEME_LOCAL_FLONUM:
-    break;
-  default:
-    flags  = SCHEME_LOCAL_OTHER_CLEARS;
-    break;
-  }
+  if ((flags < 0) || (flags > (SCHEME_MAX_LOCAL_TYPE + SCHEME_LOCAL_TYPE_OFFSET)))
+    flags = SCHEME_LOCAL_OTHER_CLEARS;
 
   if (pos < MAX_CONST_LOCAL_POS) {
     return scheme_local[pos][k][flags];
@@ -920,7 +923,8 @@ static Scheme_Local *get_frame_loc(Scheme_Comp_Env *frame,
 
 Scheme_Object *scheme_hash_module_variable(Scheme_Env *env, Scheme_Object *modidx, 
 					   Scheme_Object *stxsym, Scheme_Object *insp,
-					   int pos, intptr_t mod_phase, int is_constant)
+					   int pos, intptr_t mod_phase, int is_constant,
+                                           Scheme_Object *shape)
 /* is_constant == 2 => constant over all instantiations and phases */
 {
   Scheme_Object *val;
@@ -956,6 +960,7 @@ Scheme_Object *scheme_hash_module_variable(Scheme_Env *env, Scheme_Object *modid
       mv->insp = insp;
       mv->pos = pos;
       mv->mod_phase = (int)mod_phase;
+      mv->shape = shape;
 
       if (is_constant > 1)
         SCHEME_MODVAR_FLAGS(mv) |= SCHEME_MODVAR_CONST;
@@ -1080,61 +1085,76 @@ Scheme_Object *scheme_tl_id_sym(Scheme_Env *env, Scheme_Object *id, Scheme_Objec
     marks = SCHEME_CAR(marks);
   }
 
-  if (!SCHEME_TRUEP(bdg))
+  if (SCHEME_FALSEP(bdg))
     bdg = NULL;
 
-  /* Find a mapping that matches the longest tail of marks */
-  for (l = map; SCHEME_PAIRP(l); l = SCHEME_CDR(l)) {
-    a = SCHEME_CAR(l);
-    amarks = SCHEME_CAR(a);
+  /* Find a mapping that matches the longest tail of marks
+     in the first matching tail of bdg */
+  while (1) {
+    for (l = map; SCHEME_PAIRP(l); l = SCHEME_CDR(l)) {
+      a = SCHEME_CAR(l);
+      amarks = SCHEME_CAR(a);
 
-    if (SCHEME_VECTORP(amarks)) {
-      abdg = SCHEME_VEC_ELS(amarks)[1];
-      amarks = SCHEME_VEC_ELS(amarks)[0];
-    } else
-      abdg = NULL;
+      if (SCHEME_VECTORP(amarks)) {
+        abdg = SCHEME_VEC_ELS(amarks)[1];
+        amarks = SCHEME_VEC_ELS(amarks)[0];
+      } else
+        abdg = NULL;
 
-    if (SAME_OBJ(abdg, bdg)) {
-      if (mode > 0) {
-	if (scheme_equal(amarks, marks)) {
-	  best_match = SCHEME_CDR(a);
-	  break;
-	}
-      } else {
-        if (SCHEME_NULLP(amarks)) {
-          /* can always match empty marks */
-          best_match = SCHEME_CDR(a);
-          best_match_skipped = 0;
-        } else if (!SCHEME_PAIRP(marks)) {
-	  /* To be better than nothing, could only match exactly: */
-	  if (scheme_equal(amarks, marks)) {
-	    best_match = SCHEME_CDR(a);
-	    best_match_skipped = 0;
-	  }
-	} else {
-	  /* amarks can match a tail of marks: */
-	  for (m = marks, ms = 0; 
-	       SCHEME_PAIRP(m) && (ms < best_match_skipped);
-	       m = SCHEME_CDR(m), ms++) {
+      if (SAME_OBJ(abdg, bdg) 
+          || (bdg && abdg && scheme_equal(abdg, bdg))) {
+        if (mode > 0) {
+          if (scheme_equal(amarks, marks)) {
+            best_match = SCHEME_CDR(a);
+            break;
+          }
+        } else {
+          if (SCHEME_NULLP(amarks)) {
+            /* can always match empty marks */
+            best_match = SCHEME_CDR(a);
+            best_match_skipped = scheme_proper_list_length(marks);
+          } else if (!SCHEME_PAIRP(marks)) {
+            /* To be better than nothing, could only match exactly: */
+            if (scheme_equal(amarks, marks)) {
+              best_match = SCHEME_CDR(a);
+              best_match_skipped = 0;
+            }
+          } else {
+            /* amarks can match a tail of marks: */
+            for (m = marks, ms = 0; 
+                 SCHEME_PAIRP(m) && (ms < best_match_skipped);
+                 m = SCHEME_CDR(m), ms++) {
 
-	    cm = m;
-	    if (!SCHEME_PAIRP(amarks)) {
-	      /* If we're down to the last element
-		 of marks, then extract it to try to
-		 match the symbol amarks. */
-	      if (SCHEME_NULLP(SCHEME_CDR(m)))
-		cm = SCHEME_CAR(m);
-	    }
+              cm = m;
+              if (!SCHEME_PAIRP(amarks)) {
+                /* If we're down to the last element
+                   of marks, then extract it to try to
+                   match the symbol amarks. */
+                if (SCHEME_NULLP(SCHEME_CDR(m)))
+                  cm = SCHEME_CAR(m);
+              }
   
-	    if (scheme_equal(amarks, cm)) {
-	      best_match = SCHEME_CDR(a);
-	      best_match_skipped = ms;
-	      break;
-	    }
-	  }
-	}
+              if (scheme_equal(amarks, cm)) {
+                best_match = SCHEME_CDR(a);
+                best_match_skipped = ms;
+                break;
+              }
+            }
+          }
+        }
       }
     }
+
+    if (!best_match && (mode <= 1) && bdg && (SCHEME_PAIRP(bdg) || SCHEME_INTP(bdg) || SCHEME_BIGNUMP(bdg))) {
+      /* try lookup with less bdg context */
+      if (SCHEME_PAIRP(bdg)) {
+        bdg = SCHEME_CDR(bdg);
+        if (SCHEME_PAIRP(bdg) && SCHEME_NULLP(SCHEME_CDR(bdg)))
+          bdg = SCHEME_CAR(bdg);
+      } else
+        bdg = NULL;
+    } else
+      break;
   }
 
   if (!best_match) {
@@ -1145,15 +1165,20 @@ Scheme_Object *scheme_tl_id_sym(Scheme_Env *env, Scheme_Object *id, Scheme_Objec
     /* Last chance before making up a new name. If we're processing a
        module body generated by `expand', then we picked a name last
        time around. We can't pick a new name now, otherwise
-       "redundant" module renamings wouldn't be redundant. (See
-       simpify in "syntax.c".) So check for a context-determined
-       existing rename. */
-    if (!SCHEME_HASHTP((Scheme_Object *)env) && env->module && (mode < 2)) {
-      Scheme_Object *mod, *nm = id;
-      mod = scheme_stx_module_name(NULL, &nm, scheme_make_integer(env->phase), NULL, NULL, NULL, 
-                                   NULL, NULL, NULL, NULL, NULL);
-      if (mod /* must refer to env->module, otherwise there would
-		 have been an error before getting here */
+       "redundant" module renamings wouldn't be redundant (see
+       simpify in "syntax.c") and submodules won't re-expand correctly.
+       So, check for a context-determined existing rename. */
+    if (!SCHEME_HASHTP((Scheme_Object *)env) && env->module && (mode <= 2)) {
+      Scheme_Object *mod, *nm = id, *nom_modix = scheme_false;
+      int skipped;
+      mod = scheme_stx_module_name(NULL, &nm, scheme_make_integer(env->phase), &nom_modix, NULL, NULL, 
+                                   NULL, NULL, NULL, NULL, NULL, &skipped);
+      if (mod
+          && !SAME_OBJ(mod, scheme_undefined)
+          /* refers to env->module if nom_modix has #f path */
+          && (!SAME_TYPE(SCHEME_TYPE(nom_modix), scheme_module_index_type)
+              || SCHEME_FALSEP(((Scheme_Modidx *)nom_modix)->path))
+          && ((skipped == 0) || (mode < 2))
 	  && NOT_SAME_OBJ(nm, sym))
 	/* It has a rename already! */
 	best_match = nm;
@@ -1203,7 +1228,6 @@ Scheme_Object *scheme_tl_id_sym(Scheme_Env *env, Scheme_Object *id, Scheme_Objec
 	    break;
 	  }
 	}
-
 	/* Otherwise, increment counter and try again... */
       }
     }
@@ -1647,6 +1671,11 @@ static void check_taint(Scheme_Object *find_id)
                         "cannot use identifier tainted by macro transformation");
 }
 
+static Scheme_Object *intern_struct_proc_shape(int shape) {
+  char buf[20];
+  sprintf(buf, "struct%d", shape);
+  return scheme_intern_symbol(buf);
+}
 
 /*********************************************************************/
 /* 
@@ -1681,7 +1710,7 @@ scheme_lookup_binding(Scheme_Object *find_id, Scheme_Comp_Env *env, int flags,
   int j = 0, p = 0, modpos, skip_stops = 0, module_self_reference = 0, is_constant;
   Scheme_Bucket *b;
   Scheme_Object *val, *modidx, *modname, *src_find_id, *find_global_id, *mod_defn_phase;
-  Scheme_Object *find_id_sym = NULL, *rename_insp = NULL, *mod_constant = NULL;
+  Scheme_Object *find_id_sym = NULL, *rename_insp = NULL, *mod_constant = NULL, *shape;
   Scheme_Env *genv;
   intptr_t phase;
 
@@ -1817,7 +1846,7 @@ scheme_lookup_binding(Scheme_Object *find_id, Scheme_Comp_Env *env, int flags,
 
   src_find_id = find_id;
   modidx = scheme_stx_module_name(NULL, &find_id, scheme_make_integer(phase), NULL, NULL, &mod_defn_phase, 
-                                  NULL, NULL, NULL, NULL, &rename_insp);
+                                  NULL, NULL, NULL, NULL, &rename_insp, NULL);
 
   /* If modidx and modidx is not #<undefined>,  then find_id is now a 
      symbol, otherwise it's still an identifier. */
@@ -1857,38 +1886,37 @@ scheme_lookup_binding(Scheme_Object *find_id, Scheme_Comp_Env *env, int flags,
       genv = scheme_module_access(modname, env->genv, SCHEME_INT_VAL(mod_defn_phase));
 
       if (!genv) {
-	if (env->genv->phase) {
-	  /* The failure might be due a laziness in required-syntax
-	     execution. Force all laziness at the prior level 
-	     and try again. */
-	  scheme_module_force_lazy(env->genv, 1);
-	  genv = scheme_module_access(modname, env->genv, SCHEME_INT_VAL(mod_defn_phase));
-	}
-
-	if (!genv) {
-	  scheme_wrong_syntax("require", NULL, src_find_id, 
-                              "namespace mismatch; reference (phase %d) to a module"
-                              " %D that is not available (phase level %d)", 
-			      env->genv->phase, modname, SCHEME_INT_VAL(mod_defn_phase));
-	  return NULL;
-	}
+        scheme_wrong_syntax("require", NULL, src_find_id,
+                            "namespace mismatch;\n"
+                            " reference to a module that is not available\n"
+                            "  reference phase: %d\n"
+                            "  referenced module: %D\n"
+                            "  referenced phase level: %d",
+                            env->genv->phase, modname, SCHEME_INT_VAL(mod_defn_phase));
       }
     }
   } else {
     genv = env->genv;
     modname = NULL;
 
-    if (genv->module && (genv->disallow_unbound > 0)) {
-      /* Free identifier. Maybe don't continue. */
-      if (flags & (SCHEME_SETTING | SCHEME_REFERENCING)) {
-        scheme_unbound_syntax(((flags & SCHEME_SETTING) 
-                               ? scheme_set_stx_string
-                               : scheme_var_ref_string),
-                              NULL, src_find_id, "unbound identifier in module");
-	return NULL;
+    if (genv->module && genv->disallow_unbound) {
+      if (genv->disallow_unbound > 0) {
+        /* Free identifier. Maybe don't continue. */
+        if (flags & (SCHEME_SETTING | SCHEME_REFERENCING)) {
+          scheme_unbound_syntax(((flags & SCHEME_SETTING) 
+                                 ? scheme_set_stx_string
+                                 : scheme_var_ref_string),
+                                NULL, src_find_id, "unbound identifier in module");
+          return NULL;
+        }
+        if (flags & SCHEME_NULL_FOR_UNBOUND)
+          return NULL;
+      } else {
+        if (flags & (SCHEME_SETTING | SCHEME_REFERENCING)) {
+          scheme_register_unbound_toplevel(env, src_find_id);
+        }
+        /* continue, for now */
       }
-      if (flags & SCHEME_NULL_FOR_UNBOUND)
-	return NULL;
     }
   }
 
@@ -1930,7 +1958,7 @@ scheme_lookup_binding(Scheme_Object *find_id, Scheme_Comp_Env *env, int flags,
       pos = 0;
     else
       pos = scheme_check_accessible_in_module(genv, env->insp, in_modidx, 
-					      find_id, src_find_id, NULL, NULL, rename_insp, -1, 1,
+					      find_id, src_find_id, NULL, env->insp, rename_insp, -1, 1,
 					      _protected, NULL, env->genv, NULL, &mod_constant);
     modpos = (int)SCHEME_INT_VAL(pos);
   } else
@@ -1966,7 +1994,8 @@ scheme_lookup_binding(Scheme_Object *find_id, Scheme_Comp_Env *env, int flags,
         check_taint(src_find_id);
 	return scheme_hash_module_variable(genv, genv->module->self_modidx, find_id, 
 					   genv->module->insp,
-					   -1, genv->mod_phase, 0);
+					   -1, genv->mod_phase, 0,
+                                           NULL);
       }
     } else
       return NULL;
@@ -1974,15 +2003,25 @@ scheme_lookup_binding(Scheme_Object *find_id, Scheme_Comp_Env *env, int flags,
 
   check_taint(src_find_id);
 
+  shape = NULL;
   if (mod_constant) {
     if (SAME_OBJ(mod_constant, scheme_constant_key))
       is_constant = 2;
     else if (SAME_OBJ(mod_constant, scheme_fixed_key))
       is_constant = 1;
-    else if (SAME_TYPE(SCHEME_TYPE(mod_constant), scheme_inline_variant_type)) {
+    else if (SAME_TYPE(SCHEME_TYPE(mod_constant), scheme_proc_shape_type)) {
+      is_constant = 2;
+      shape = SCHEME_PTR_VAL(mod_constant);
+    } else if (SAME_TYPE(SCHEME_TYPE(mod_constant), scheme_struct_proc_shape_type)) {
       if (_inline_variant)
         *_inline_variant = mod_constant;
       is_constant = 2;
+      shape = intern_struct_proc_shape(SCHEME_PROC_SHAPE_MODE(mod_constant));
+    } else if (SAME_TYPE(SCHEME_TYPE(mod_constant), scheme_inline_variant_type)) {
+      if (_inline_variant)
+        *_inline_variant = mod_constant;
+      is_constant = 2;
+      shape = scheme_get_or_check_procedure_shape(mod_constant, NULL);
     } else {
       if (flags & SCHEME_ELIM_CONST) 
         return mod_constant;
@@ -1997,13 +2036,14 @@ scheme_lookup_binding(Scheme_Object *find_id, Scheme_Comp_Env *env, int flags,
       && (!(scheme_is_kernel_modname(modname) 
             || scheme_is_unsafe_modname(modname)
             || scheme_is_flfxnum_modname(modname)
+            || scheme_is_extfl_modname(modname)
             || scheme_is_futures_modname(modname))
           || (flags & SCHEME_REFERENCING))) {
     /* Create a module variable reference, so that idx is preserved: */
     return scheme_hash_module_variable(env->genv, modidx, find_id, 
 				       (rename_insp ? rename_insp : genv->module->insp),
 				       modpos, SCHEME_INT_VAL(mod_defn_phase),
-                                       is_constant);
+                                       is_constant, shape);
   }
 
   if (!modname 
@@ -2014,7 +2054,7 @@ scheme_lookup_binding(Scheme_Object *find_id, Scheme_Comp_Env *env, int flags,
     return scheme_hash_module_variable(env->genv, genv->module->self_modidx, find_global_id, 
 				       genv->module->insp,
 				       modpos, genv->mod_phase,
-                                       is_constant);
+                                       is_constant, shape);
   }
 
   b = scheme_bucket_from_table(genv->toplevel, (char *)find_global_id);
@@ -2071,6 +2111,16 @@ Scheme_Object *scheme_extract_flfxnum(Scheme_Object *o)
     return NULL;
 }
 
+Scheme_Object *scheme_extract_extfl(Scheme_Object *o)
+{
+  Scheme_Env *home;
+  home = scheme_get_bucket_home((Scheme_Bucket *)o);
+  if (home && home->module && scheme_is_extfl_modname(home->module->modname))
+    return (Scheme_Object *)((Scheme_Bucket *)o)->val;
+  else
+    return NULL;
+}
+
 Scheme_Object *scheme_extract_futures(Scheme_Object *o)
 {
   Scheme_Env *home;
@@ -2150,21 +2200,21 @@ scheme_do_local_lift_expr(const char *who, int stx_pos, int argc, Scheme_Object
       count = -1;
 
     if (count < 0)
-      scheme_wrong_type(who, "exact nonnegative integer", 0, argc, argv);
+      scheme_wrong_contract(who, "exact-nonnegative-integer?", 0, argc, argv);
   } else
     count = 1;
 
   expr = argv[stx_pos];
   if (!SCHEME_STXP(expr))
-    scheme_wrong_type(who, "syntax", stx_pos, argc, argv);
+    scheme_wrong_contract(who, "syntax?", stx_pos, argc, argv);
 
   env = orig_env = scheme_current_thread->current_local_env;
   local_mark = scheme_current_thread->current_local_mark;
 
   if (!env)
-    scheme_raise_exn(MZEXN_FAIL_CONTRACT, 
-		     "%s: not currently transforming",
-                     who);
+    scheme_contract_error(who,
+                          "not currently transforming",
+                          NULL);
 
   while (env && !COMPILE_DATA(env)->lifts) {
     env = env->next;
@@ -2175,8 +2225,9 @@ scheme_do_local_lift_expr(const char *who, int stx_pos, int argc, Scheme_Object
       env = NULL;
 
   if (!env)
-    scheme_raise_exn(MZEXN_FAIL_CONTRACT, 
-		     "syntax-local-lift-expression: no lift target");
+    scheme_contract_error("syntax-local-lift-expression",
+                          "no lift target",
+                          NULL);
   
   expr = scheme_add_remove_mark(expr, local_mark);
 
@@ -2254,9 +2305,10 @@ scheme_local_lift_end_statement(Scheme_Object *expr, Scheme_Object *local_mark,
   env = scheme_get_module_lift_env(env);
 
   if (!env)
-    scheme_raise_exn(MZEXN_FAIL_CONTRACT, 
-		     "syntax-local-lift-module-end-declaration: not currently transforming"
-                     " an expression within a module declaration");
+    scheme_contract_error("syntax-local-lift-module-end-declaration",
+                          "not currently transforming"
+                          " an expression within a module declaration",
+                          NULL);
   
   expr = scheme_add_remove_mark(expr, local_mark);
   orig_expr = expr;
@@ -2291,8 +2343,9 @@ Scheme_Object *scheme_local_lift_require(Scheme_Object *form, Scheme_Object *ori
   }
 
   if (!env)
-    scheme_raise_exn(MZEXN_FAIL_CONTRACT, 
-		     "syntax-local-lift-requires: could not find target context");
+    scheme_contract_error("syntax-local-lift-requires",
+                          "could not find target context",
+                          NULL);
 
   
   mark = scheme_new_mark();
@@ -2314,6 +2367,10 @@ Scheme_Object *scheme_local_lift_require(Scheme_Object *form, Scheme_Object *ori
 
   SCHEME_EXPAND_OBSERVE_LIFT_REQUIRE(scheme_get_expand_observe(), req_form, orig_form, form);
 
+  /* In a top-level context, may need to force compile-time evaluation: */
+  if (!env->genv->module)
+    scheme_prepare_compile_env(env->genv);
+
   return form;
 }
 
@@ -2331,8 +2388,9 @@ Scheme_Object *scheme_local_lift_provide(Scheme_Object *form, Scheme_Object *loc
   }
 
   if (!env)
-    scheme_raise_exn(MZEXN_FAIL_CONTRACT, 
-		     "syntax-local-lift-provide: not expanding in a module run-time body");
+    scheme_contract_error("syntax-local-lift-provide",
+                          "not expanding in a module run-time body",
+                          NULL);
   
   form = scheme_add_remove_mark(form, local_mark);
   form = scheme_datum_to_syntax(scheme_make_pair(scheme_datum_to_syntax(scheme_intern_symbol("#%provide"), 
@@ -2386,11 +2444,6 @@ Scheme_Object *scheme_find_local_shadower(Scheme_Object *sym, Scheme_Object *sym
   Scheme_Comp_Env *frame;
   Scheme_Object *esym, *uid = NULL, *env_marks, *prop;
 
-  if (!unshadowable_symbol) {
-    REGISTER_SO(unshadowable_symbol);
-    unshadowable_symbol = scheme_intern_symbol("unshadowable");
-  }
-
   /* Walk backward through the frames, looking for a renaming binding
      with the same marks as the given identifier, sym. Skip over
      unsealed ribs, though. When we find a match, rename the given
