@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2012 PLT Scheme Inc.
+  Copyright (c) 2004-2013 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -300,6 +300,7 @@ Scheme_Object *scheme_complex_power(const Scheme_Object *base, const Scheme_Obje
   Scheme_Complex *cb = (Scheme_Complex *)base;
   Scheme_Complex *ce = (Scheme_Complex *)exponent;
   double a, b, c, d, bm, ba, nm, na, r1, r2;
+  int d_is_zero;
 
   if ((ce->i == zero) && !SCHEME_FLOATP(ce->r)) {
     if (SCHEME_INTP(ce->r) || SCHEME_BIGNUMP(ce->r))
@@ -310,28 +311,31 @@ Scheme_Object *scheme_complex_power(const Scheme_Object *base, const Scheme_Obje
   b = scheme_get_val_as_double(cb->i);
   c = scheme_get_val_as_double(ce->r);
   d = scheme_get_val_as_double(ce->i);
+  d_is_zero = (ce->i == zero);
 
   bm = sqrt(a * a + b * b);
   ba = atan2(b, a);
 
   /* New mag & angle */
-  nm = pow(bm, c) * exp(-(ba * d));
-  na = log(bm) * d + ba * c;
+  nm = scheme_double_expt(bm, c) * exp(-(ba * d));
+  if (d_is_zero) /* precision here can avoid NaNs */
+    na = ba * c;
+  else
+    na = log(bm) * d + ba * c;
 
   r1 = nm * cos(na);
   r2 = nm * sin(na);
 
 #ifdef MZ_USE_SINGLE_FLOATS
   /* Coerce to double or float? */
-#ifdef USE_SINGLE_FLOATS_AS_DEFAULT
   if (!SCHEME_DBLP(cb->r) && !SCHEME_DBLP(cb->i)
       && !SCHEME_DBLP(ce->r) && !SCHEME_DBLP(ce->i))
-#else
-  if (SCHEME_FLTP(cb->r) && SCHEME_FLTP(cb->i)
-      && SCHEME_FLTP(ce->r) && SCHEME_FLTP(ce->i))
+#ifndef USE_SINGLE_FLOATS_AS_DEFAULT
+    if (SCHEME_FLTP(cb->r) || SCHEME_FLTP(cb->i)
+        || SCHEME_FLTP(ce->r) || SCHEME_FLTP(ce->i))
 #endif
-    return scheme_make_complex(scheme_make_float((float)r1), 
-			       scheme_make_float((float)r2));
+      return scheme_make_complex(scheme_make_float((float)r1), 
+                                 scheme_make_float((float)r2));
 #endif
 
   return scheme_make_complex(scheme_make_double(r1), 
@@ -351,8 +355,20 @@ Scheme_Object *scheme_complex_sqrt(const Scheme_Object *o)
     r = scheme_sqrt(1, &r);
     if (!SCHEME_COMPLEXP(r))
       return scheme_make_complex(r, i);
-    else
-      return r;
+    else {
+      c = (Scheme_Complex *)r;
+      if (SAME_OBJ(c->r, zero)) {
+        /* need an inexact-zero real part: */
+#ifdef MZ_USE_SINGLE_FLOATS
+        if (SCHEME_FLTP(c->i))
+          r = scheme_make_float(0.0);
+        else
+#endif
+          r = scheme_make_double(0.0);
+        return scheme_make_complex(r, c->i);
+      } else
+        return r;
+    }
   }
 
   ssq = scheme_bin_plus(scheme_bin_mult(r, r),
@@ -363,9 +379,15 @@ Scheme_Object *scheme_complex_sqrt(const Scheme_Object *o)
   if (SCHEME_FLOATP(srssq)) {
     /* We may have lost too much precision, if i << r.  The result is
        going to be inexact, anyway, so switch to using expt. */
-    Scheme_Object *a[2];
+    Scheme_Object *a[2], *p;
     a[0] = (Scheme_Object *)o;
-    a[1] = scheme_make_double(0.5);
+#ifdef MZ_USE_SINGLE_FLOATS
+    if (SCHEME_FLTP(c->i))
+      p = scheme_make_float(0.5);
+    else
+#endif
+      p = scheme_make_double(0.5);
+    a[1] = p;
     return scheme_expt(2, a);
   }
 
