@@ -7,7 +7,8 @@
      visible to the GC.
 
   3) Immediate operands must be 32-bit values on x86_64, except with
-     jit_movi, jit_sti, jit_ldi, jit_bXi, jit_calli, and jit_finishi.
+     jit_movi, jit_sti, jit_ldi, jit_bXi, jit_calli (in default-long
+     mode), and jit_finishi.
 
   4) Function calls are limited to 3 arguments (i.e., jit_prepare()
      must never be called with a number greater than 3). This limit
@@ -97,12 +98,32 @@ END_XFORM_ARITH;
 #define JIT_LOG_DOUBLE_SIZE 3
 #define JIT_DOUBLE_SIZE (1 << JIT_LOG_DOUBLE_SIZE)
 
+#ifdef MZ_LONG_DOUBLE
+# ifdef MZ_USE_JIT_X86_64
+#  define JIT_LOG_LONG_DOUBLE_SIZE 4
+#  define JIT_LONG_DOUBLE_SIZE (1 << JIT_LOG_LONG_DOUBLE_SIZE)
+# else
+#  define JIT_LOG_LONG_DOUBLE_SIZE not_implemented
+#  define JIT_LONG_DOUBLE_SIZE 12
+#endif
+
+#endif
+
 /* a mzchar is an int: */
 #define LOG_MZCHAR_SIZE 2
 
 #if defined(MZ_USE_JIT_PPC) || defined(MZ_USE_JIT_X86_64)
+/* Both PPC and x86_64 need long jumps, sometimes */
 # define NEED_LONG_JUMPS
 #endif
+#if defined(MZ_USE_JIT_PPC)
+/* For PPC, long jumps may be needed even within a JIT-generated block */
+# define NEED_LONG_BRANCHES
+#endif
+#if defined(MZ_USE_JIT_X86_64)
+/* For x86_64, long jumps are needed only if we start allocating far away */
+# define SET_DEFAULT_LONG_JUMPS
+#endif
 /* Tiny jumps seem worthwhile for x86, but they don't seem to help for x86_64: */
 #if defined(MZ_USE_JIT_I386) && !defined(MZ_USE_JIT_X86_64)
 # define USE_TINY_JUMPS
@@ -141,6 +162,8 @@ Fix me! See use.
 # endif
 #endif
 
+#include "jitfpu.h"
+
 #if 0
 static void assert_failure(int where) { printf("JIT assert failed %d\n", where); }
 #define JIT_ASSERT(v) if (!(v)) assert_failure(__LINE__);
@@ -173,6 +196,11 @@ extern int scheme_direct_call_count, scheme_indirect_call_count;
 extern int scheme_jit_malloced;
 #ifdef JIT_USE_FP_OPS
 THREAD_LOCAL_DECL(extern double scheme_jit_save_fp);
+THREAD_LOCAL_DECL(extern double scheme_jit_save_fp2);
+# ifdef MZ_LONG_DOUBLE
+THREAD_LOCAL_DECL(extern long_double scheme_jit_save_extfp);
+THREAD_LOCAL_DECL(extern long_double scheme_jit_save_extfp2);
+# endif
 #endif
 
 typedef int (*Native_Check_Arity_Proc)(Scheme_Object *o, int argc, int dummy EXTRA_NATIVE_ARGUMENT_TYPE);
@@ -204,14 +232,24 @@ typedef struct Apply_LWC_Args {
 typedef Scheme_Object *(*Continuation_Apply_Indirect)(Apply_LWC_Args *, intptr_t);
 typedef Scheme_Object *(*Continuation_Apply_Finish)(Apply_LWC_Args *args, void *stack, void *frame);
 
+#ifdef MZ_LONG_DOUBLE
+# define JIT_NUM_FL_KINDS 2
+#else
+# define JIT_NUM_FL_KINDS 1
+#endif
+
 struct scheme_jit_common_record {
   int skip_checks;
 
 #define MAX_SHARED_CALL_RANDS 25
   void *shared_tail_code[4][MAX_SHARED_CALL_RANDS];
-  void *shared_non_tail_code[4][MAX_SHARED_CALL_RANDS][2];
-  void *shared_non_tail_retry_code[2];
-  void *shared_non_tail_argc_code[2];
+# define SHARED_SINGLE_VALUE_CASE 0
+# define SHARED_MULTI_OK_CASE 1
+# define SHARED_RESULT_IGNORED_CASE 2
+# define SHARED_NUM_NONTAIL_CASES 3
+  void *shared_non_tail_code[5][MAX_SHARED_CALL_RANDS][SHARED_NUM_NONTAIL_CASES];
+  void *shared_non_tail_retry_code[SHARED_NUM_NONTAIL_CASES];
+  void *shared_non_tail_argc_code[SHARED_NUM_NONTAIL_CASES];
   void *shared_tail_argc_code;
 
 #define MAX_SHARED_ARITY_CHECK 25
@@ -234,7 +272,7 @@ struct scheme_jit_common_record {
   void *bad_set_mcar_code, *bad_set_mcdr_code;
   void *imag_part_code, *real_part_code, *make_rectangular_code;
   void *bad_flimag_part_code, *bad_flreal_part_code, *bad_make_flrectangular_code;
-  void *unbox_code, *set_box_code;
+  void *unbox_code, *set_box_code, *box_cas_fail_code;
   void *bad_vector_length_code;
   void *bad_flvector_length_code;
   void *bad_fxvector_length_code;
@@ -242,23 +280,28 @@ struct scheme_jit_common_record {
   void *chap_vector_ref_code, *chap_vector_ref_check_index_code, *chap_vector_set_code, *chap_vector_set_check_index_code;
   void *string_ref_code, *string_ref_check_index_code, *string_set_code, *string_set_check_index_code;
   void *bytes_ref_code, *bytes_ref_check_index_code, *bytes_set_code, *bytes_set_check_index_code;
-  void *flvector_ref_check_index_code, *flvector_set_check_index_code, *flvector_set_flonum_check_index_code;
+  void *flvector_ref_check_index_code[JIT_NUM_FL_KINDS];
+  void *flvector_set_check_index_code[JIT_NUM_FL_KINDS], *flvector_set_flonum_check_index_code[JIT_NUM_FL_KINDS];
   void *fxvector_ref_code, *fxvector_ref_check_index_code, *fxvector_set_code, *fxvector_set_check_index_code;
   void *struct_raw_ref_code, *struct_raw_set_code;
   void *syntax_e_code;
   void *on_demand_jit_arity_code, *in_progress_on_demand_jit_arity_code;
   void *get_stack_pointer_code;
   void *stack_cache_pop_code;
-  void *struct_pred_code, *struct_pred_multi_code;
+  void *struct_pred_code, *struct_pred_tail_code, *struct_pred_multi_code;
   void *struct_pred_branch_code;
-  void *struct_get_code, *struct_get_multi_code;
-  void *struct_set_code, *struct_set_multi_code;
-  void *struct_prop_get_code, *struct_prop_get_multi_code;
-  void *struct_prop_get_defl_code, *struct_prop_get_defl_multi_code;
-  void *struct_prop_pred_code, *struct_prop_pred_multi_code;
+  void *struct_get_code, *struct_get_tail_code, *struct_get_multi_code;
+  void *struct_set_code, *struct_set_tail_code, *struct_set_multi_code;
+  void *struct_prop_get_code, *struct_prop_get_tail_code, *struct_prop_get_multi_code;
+  void *struct_prop_get_defl_code, *struct_prop_get_defl_tail_code, *struct_prop_get_defl_multi_code;
+  void *struct_prop_pred_code, *struct_prop_pred_tail_code, *struct_prop_pred_multi_code;
   void *struct_proc_extract_code;
+  void *struct_constr_unary_code, *struct_constr_unary_tail_code, *struct_constr_unary_multi_code;
+  void *struct_constr_binary_code, *struct_constr_binary_tail_code, *struct_constr_binary_multi_code;
+  void *struct_constr_nary_code, *struct_constr_nary_tail_code, *struct_constr_nary_multi_code;
   void *bad_app_vals_target;
   void *app_values_slow_code, *app_values_multi_slow_code, *app_values_tail_slow_code;
+  void *bad_char_to_integer_code, *slow_integer_to_char_code;
   void *values_code;
   void *list_p_code, *list_p_branch_code;
   void *list_length_code;
@@ -266,8 +309,13 @@ struct scheme_jit_common_record {
   void *finish_tail_call_code, *finish_tail_call_fixup_code;
   void *module_run_start_code, *module_exprun_start_code, *module_start_start_code;
   void *box_flonum_from_stack_code, *box_flonum_from_reg_code;
-  void *fl1_fail_code, *fl2rr_fail_code[2], *fl2fr_fail_code[2], *fl2rf_fail_code[2];
-  void *wcm_code, *wcm_nontail_code;
+  void *fl1_fail_code[JIT_NUM_FL_KINDS], *fl2rr_fail_code[2][JIT_NUM_FL_KINDS];
+  void *fl2fr_fail_code[2][JIT_NUM_FL_KINDS], *fl2rf_fail_code[2][JIT_NUM_FL_KINDS];
+#ifdef MZ_LONG_DOUBLE
+  void *bad_extflvector_length_code;
+  void *box_extflonum_from_stack_code, *box_extflonum_from_reg_code;
+#endif
+  void *wcm_code, *wcm_nontail_code, *wcm_chaperone;
   void *apply_to_list_tail_code, *apply_to_list_code, *apply_to_list_multi_ok_code;
   void *eqv_code, *eqv_branch_code;
   void *proc_arity_includes_code;
@@ -277,6 +325,9 @@ struct scheme_jit_common_record {
   void *retry_alloc_code;
   void *retry_alloc_code_keep_r0_r1;
   void *retry_alloc_code_keep_fpr1;
+# ifdef MZ_LONG_DOUBLE
+  void *retry_alloc_code_keep_extfpr1;
+# endif
 #endif
 
   Continuation_Apply_Indirect continuation_apply_indirect_code;
@@ -311,7 +362,7 @@ typedef struct mz_jit_state {
                     .             shift >>2 to get flags
 		    . 1 -> case 0x2 bit:
                     .        0 -> shift >>2 to get new (native) pushed
-                    .        1 -> shift >>2 to get flonum stack pos */
+                    .        1 -> shift >>2 to get flostack offset */
   int num_mappings, mappings_size;
   int retained, retained_double;
   int need_set_rs;
@@ -335,6 +386,9 @@ typedef struct mz_jit_state {
   int rs_virtual_offset;
   int unbox, unbox_depth;
   int flostack_offset, flostack_space;
+#ifdef MZ_LONG_DOUBLE
+  int unbox_extflonum;
+#endif
   int self_restart_offset, self_restart_space;
 } mz_jit_state;
 
@@ -396,6 +450,11 @@ typedef struct {
 # define tl_fixup_runstack_base            tl_delta(fixup_runstack_base)
 # define tl_fixup_already_in_place         tl_delta(fixup_already_in_place)
 # define tl_scheme_jit_save_fp             tl_delta(scheme_jit_save_fp)
+# define tl_scheme_jit_save_fp2            tl_delta(scheme_jit_save_fp2)
+#ifdef MZ_LONG_DOUBLE
+# define tl_scheme_jit_save_extfp          tl_delta(scheme_jit_save_extfp)
+# define tl_scheme_jit_save_extfp2         tl_delta(scheme_jit_save_extfp2)
+#endif
 # define tl_scheme_fuel_counter            tl_delta(scheme_fuel_counter)
 # define tl_scheme_jit_stack_boundary      tl_delta(scheme_jit_stack_boundary)
 # define tl_jit_future_storage             tl_delta(jit_future_storage)
@@ -408,13 +467,23 @@ void *scheme_jit_get_threadlocal_table();
 # ifdef JIT_X86_64
 #  define JIT_R10 JIT_R(10)
 #  define JIT_R14 JIT_R(14)
-#  define mz_tl_addr(reg, addr) LEAQmQr((addr), (JIT_R14), 0, 0, (reg))
-#  define mz_tl_addr_tmp(tmp_reg, addr) (mz_tl_addr(JIT_R10, addr))
+#  define mz_tl_addr(reg, addr) (void)0
+#  define mz_tl_addr_tmp(tmp_reg, addr) (void)0
 #  define mz_tl_addr_untmp(tmp_reg) (void)0
 #  define mz_tl_tmp_reg(tmp_reg) JIT_R10
-#  define _mz_tl_str_p(addr, tmp_reg, reg) jit_str_p(tmp_reg, reg)
-#  define _mz_tl_str_l(addr, tmp_reg, reg) jit_str_l(tmp_reg, reg)
-#  define _mz_tl_str_i(addr, tmp_reg, reg) jit_str_i(tmp_reg, reg)
+#  define _mz_tl_str_p(addr, tmp_reg, reg) jit_stxi_p(addr, JIT_R14, reg)
+#  define _mz_tl_str_l(addr, tmp_reg, reg) jit_stxi_l(addr, JIT_R14, reg)
+#  define _mz_tl_str_i(addr, tmp_reg, reg) jit_stxi_i(addr, JIT_R14, reg)
+#  define mz_tl_ldr_p(reg, addr) jit_ldxi_p(reg, JIT_R14, addr)
+#  define mz_tl_ldr_l(reg, addr) jit_ldxi_l(reg, JIT_R14, addr)
+#  define mz_tl_ldr_i(reg, addr) jit_ldxi_i(reg, JIT_R14, addr)
+#  define mz_tl_str_d_fppop(tmp_reg, reg, addr) jit_stxi_d_fppop(addr, JIT_R14, reg)
+#  define mz_tl_ldr_d_fppush(reg, tmp_reg, addr) jit_ldxi_d_fppush(reg, JIT_R14, addr)
+#  define mz_fpu_tl_str_ld_fppop(tmp_reg, reg, addr) jit_fpu_stxi_ld_fppop(addr, JIT_R14, reg)
+#  define mz_fpu_tl_ldr_ld_fppush(reg, tmp_reg, addr) jit_fpu_ldxi_ld_fppush(reg, JIT_R14, addr)
+#  define mz_tl_addr_tmp_i(tmp_reg, addr) (void)0
+#  define mz_tl_addr_untmp_i(tmp_reg) (void)0
+#  define mz_tl_tmp_reg_i(tmp_reg) tmp_reg
 # else
 #  define THREAD_LOCAL_USES_JIT_V2
 #  ifdef THREAD_LOCAL_USES_JIT_V2
@@ -434,17 +503,31 @@ void *scheme_jit_get_threadlocal_table();
 #   define _mz_tl_str_l(addr, tmp_reg, reg) jit_str_l(tmp_reg, reg)
 #   define _mz_tl_str_i(addr, tmp_reg, reg) jit_str_i(tmp_reg, reg)
 #  endif
+#  define mz_tl_addr_tmp_i(tmp_reg, addr) mz_tl_addr_tmp(tmp_reg, addr)
+#  define mz_tl_addr_untmp_i(tmp_reg) mz_tl_addr_untmp(tmp_reg)
+#  define mz_tl_tmp_reg_i(tmp_reg) mz_tl_tmp_reg(tmp_reg)
+#  define mz_tl_ldr_p(reg, addr) jit_ldr_p(reg, reg)
+#  define mz_tl_ldr_l(reg, addr) jit_ldr_l(reg, reg)
+#  define mz_tl_ldr_i(reg, addr) jit_ldr_i(reg, reg)
+#  define mz_tl_str_d_fppop(tmp_reg, reg, addr) jit_str_d_fppop(tmp_reg, reg)
+#  define mz_tl_ldr_d_fppush(reg, tmp_reg, addr) jit_ldr_d_fppush(reg, tmp_reg)
+#  define mz_fpu_tl_str_ld_fppop(tmp_reg, reg, addr) jit_fpu_str_ld_fppop(tmp_reg, reg)
+#  define mz_fpu_tl_ldr_ld_fppush(reg, tmp_reg, addr) jit_fpu_ldr_ld_fppush(reg, tmp_reg)
 # endif
 
 /* A given tmp_reg doesn't have to be unused; it just has to be distinct from other arguments. */
 # define mz_tl_sti_p(addr, reg, tmp_reg) (mz_tl_addr_tmp(tmp_reg, addr), _mz_tl_str_p(addr, mz_tl_tmp_reg(tmp_reg), reg), mz_tl_addr_untmp(tmp_reg))
 # define mz_tl_sti_l(addr, reg, tmp_reg) (mz_tl_addr_tmp(tmp_reg, addr), _mz_tl_str_l(addr, mz_tl_tmp_reg(tmp_reg), reg), mz_tl_addr_untmp(tmp_reg))
-# define mz_tl_sti_i(addr, reg, tmp_reg) (mz_tl_addr_tmp(tmp_reg, addr), _mz_tl_str_i(addr, mz_tl_tmp_reg(tmp_reg), reg), mz_tl_addr_untmp(tmp_reg))
-# define mz_tl_ldi_p(reg, addr) (mz_tl_addr(reg, addr), jit_ldr_p(reg, reg))
-# define mz_tl_ldi_l(reg, addr) (mz_tl_addr(reg, addr), jit_ldr_l(reg, reg))
-# define mz_tl_ldi_i(reg, addr) (mz_tl_addr(reg, addr), jit_ldr_i(reg, reg))
-# define mz_tl_sti_d_fppop(addr, reg, tmp_reg) (mz_tl_addr(tmp_reg, addr), jit_str_d_fppop(tmp_reg, reg))
-# define mz_tl_ldi_d_fppush(reg, addr, tmp_reg) (mz_tl_addr(tmp_reg, addr), jit_ldr_d_fppush(reg, tmp_reg))
+# define mz_tl_sti_i(addr, reg, tmp_reg) (mz_tl_addr_tmp_i(tmp_reg, addr), _mz_tl_str_i(addr, mz_tl_tmp_reg_i(tmp_reg), reg), mz_tl_addr_untmp_i(tmp_reg))
+# define mz_tl_ldi_p(reg, addr) (mz_tl_addr(reg, addr), mz_tl_ldr_p(reg, addr))
+# define mz_tl_ldi_l(reg, addr) (mz_tl_addr(reg, addr), mz_tl_ldr_l(reg, addr))
+# define mz_tl_ldi_i(reg, addr) (mz_tl_addr(reg, addr), mz_tl_ldr_i(reg, addr))
+# define mz_tl_sti_d_fppop(addr, reg, tmp_reg) (mz_tl_addr(tmp_reg, addr), mz_tl_str_d_fppop(tmp_reg, reg, addr))
+# define mz_tl_ldi_d_fppush(reg, addr, tmp_reg) (mz_tl_addr(tmp_reg, addr), mz_tl_ldr_d_fppush(reg, tmp_reg, addr))
+# ifdef MZ_LONG_DOUBLE
+#  define mz_fpu_tl_sti_ld_fppop(addr, reg, tmp_reg) (mz_tl_addr(tmp_reg, addr), mz_fpu_tl_str_ld_fppop(tmp_reg, reg, addr))
+#  define mz_fpu_tl_ldi_ld_fppush(reg, addr, tmp_reg) (mz_tl_addr(tmp_reg, addr), mz_fpu_tl_ldr_ld_fppush(reg, tmp_reg, addr))
+# endif
 #else
 # define mz_tl_sti_p(addr, reg, tmp_reg) jit_sti_p(addr, reg)
 # define mz_tl_sti_l(addr, reg, tmp_reg) jit_sti_l(addr, reg)
@@ -454,6 +537,8 @@ void *scheme_jit_get_threadlocal_table();
 # define mz_tl_ldi_i(reg, addr) jit_ldi_i(reg, addr)
 # define mz_tl_sti_d_fppop(addr, reg, tmp_reg) jit_sti_d_fppop(addr, reg)
 # define mz_tl_ldi_d_fppush(reg, addr, tmp_reg) jit_ldi_d_fppush(reg, addr)
+# define mz_fpu_tl_sti_ld_fppop(addr, reg, tmp_reg) jit_fpu_sti_ld_fppop(addr, reg)
+# define mz_fpu_tl_ldi_ld_fppush(reg, addr, tmp_reg) jit_fpu_ldi_ld_fppush(reg, addr)
 # define tl_MZ_RUNSTACK (&MZ_RUNSTACK)
 # define tl_MZ_RUNSTACK_START (&MZ_RUNSTACK_START)
 # define tl_GC_gen0_alloc_page_ptr (&GC_gen0_alloc_page_ptr)
@@ -465,6 +550,11 @@ void *scheme_jit_get_threadlocal_table();
 # define tl_fixup_runstack_base (&fixup_runstack_base)
 # define tl_fixup_already_in_place (&fixup_already_in_place)
 # define tl_scheme_jit_save_fp (&scheme_jit_save_fp)
+# define tl_scheme_jit_save_fp2 (&scheme_jit_save_fp2)
+# ifdef MZ_LONG_DOUBLE
+#  define tl_scheme_jit_save_extfp (&scheme_jit_save_extfp)
+#  define tl_scheme_jit_save_extfp2 (&scheme_jit_save_extfp2)
+# endif
 # define tl_scheme_fuel_counter (&scheme_fuel_counter)
 # define tl_scheme_jit_stack_boundary (&scheme_jit_stack_boundary)
 #endif
@@ -529,7 +619,7 @@ static void *top4;
    mostly improves code size, rather than speed. Functions that cause
    the register to be de-sync'd are marked as such. Functions that can
    accommodate a de-sync'd register on entry are marked as such. All
-   other fuctions can assume a sync'd regsiter and ensure a sync'd
+   other functions can assume a sync'd regsiter and ensure a sync'd
    register. Note that branches and calls normally require a sync'd
    register. */
 
@@ -670,7 +760,7 @@ int check_location;
       LOCAL4 (x86_64: = saved R14 otherwise when THREAD_LOCAL
               x86: = RUNSTACK_BASE or THREAD_LOCAL)
       [some empty slots, maybe, depending on alignment]
-      [space for local, unboxed flonums]
+      [space for "flostack" --- local unboxed values, such as flonums]
     Registers: JIT_V1 = RUNSTACK, JIT_V2 = x86_64: RUNSTACK_BASE
                                            x86: RUNSTACK_BASE or THREAD_LOCAL
                x86_64: JIT_R14 = THREAD_LOCAL
@@ -680,13 +770,16 @@ int check_location;
 # define NEED_LOCAL4
 #endif
 
+#define mz_set_local_p(x, l) mz_set_local_p_x(x, l, JIT_FP)
+#define mz_get_local_p(x, l) mz_get_local_p_x(x, l, JIT_FP)
+
 #ifdef MZ_USE_JIT_PPC
 /* JIT_LOCAL1, JIT_LOCAL2, and JIT_LOCAL3 are offsets in the stack frame. */
 # define JIT_LOCAL1 56
 # define JIT_LOCAL2 60
 # define JIT_LOCAL3 64
-# define mz_set_local_p(x, l) jit_stxi_p(l, JIT_FP, x)
-# define mz_get_local_p(x, l) jit_ldxi_p(x, JIT_FP, l)
+# define mz_set_local_p_x(x, l, FP) jit_stxi_p(l, FP, x)
+# define mz_get_local_p_x(x, l, FP) jit_ldxi_p(x, FP, l)
 # define mz_patch_branch_at(a, v) (_jitl.long_jumps ? (void)jit_patch_movei(a-4, a-3, v) : (void)jit_patch_branch(a-1, v))
 # define mz_patch_ucbranch_at(a, v) (_jitl.long_jumps ? (void)jit_patch_movei(a-4, a-3, v) : (void)jit_patch_ucbranch(a-1, v))
 # define mz_prolog(x) (MFLRr(x), mz_set_local_p(x, JIT_LOCAL2))
@@ -738,8 +831,8 @@ void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
    1 word (for the return address) below alignment. */
 # define JIT_LOCAL1 -(JIT_WORD_SIZE * 4)
 # define JIT_LOCAL2 -(JIT_WORD_SIZE * 5)
-# define mz_set_local_p(x, l) jit_stxi_p((l), JIT_FP, (x))
-# define mz_get_local_p(x, l) jit_ldxi_p((x), JIT_FP, (l))
+# define mz_set_local_p_x(x, l, FP) jit_stxi_p((l), FP, (x))
+# define mz_get_local_p_x(x, l, FP) jit_ldxi_p((x), FP, (l))
 # define mz_patch_branch_at(a, v) jit_patch_branch_at(a, v)
 # define mz_patch_ucbranch_at(a, v) jit_patch_ucbranch_at(a, v)
   /* The ABI for _CALL_DARWIN or JIT_X86_64 requires alignment. Even
@@ -784,7 +877,7 @@ void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
 # endif
 # define mz_push_locals() SUBQir((LOCAL_FRAME_SIZE << JIT_LOG_WORD_SIZE), JIT_SP)
 # define mz_pop_locals() ADDQir((LOCAL_FRAME_SIZE << JIT_LOG_WORD_SIZE), JIT_SP)
-# define JIT_FRAME_FLONUM_OFFSET (-(JIT_WORD_SIZE * (LOCAL_FRAME_SIZE + 3)))
+# define JIT_FRAME_FLOSTACK_OFFSET (-(JIT_WORD_SIZE * (LOCAL_FRAME_SIZE + 3)))
 # define _jit_prolog_again(jitter, n, ret_addr_reg) (PUSHQr(ret_addr_reg), jit_base_prolog())
 # if defined(MZ_USE_JIT_X86_64) && !defined(_WIN64)
 #  define jit_shuffle_saved_regs() (MOVQrr(_ESI, _R12), MOVQrr(_EDI, _R13))
@@ -805,7 +898,7 @@ void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
 #  define mz_pop_threadlocal() /* empty */
 #  ifdef THREAD_LOCAL_USES_JIT_V2
 #   define _mz_install_threadlocal(reg) jit_movr_p(JIT_V2, reg)
-#   define mz_repush_threadlocal(in) /* empty */
+#   define mz_repush_threadlocal() /* empty */
 #  else
 #   define _mz_install_threadlocal(reg) mz_set_local_p(reg, JIT_LOCAL4)
 #   define mz_repush_threadlocal() (PUSHQr(JIT_R0), jit_ldr_p(JIT_R0, _EBP), \
@@ -823,12 +916,30 @@ void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
 # define mz_repush_threadlocal() /* empty */
 #endif
 
+#if 0
+static jit_insn *fp_tmpr;
+# define check_fp_depth(i, FP) \
+  (jit_addi_l(FP, FP, (JIT_FRAME_FLOSTACK_OFFSET - (i))),             \
+   fp_tmpr = jit_bger_l(0, FP, JIT_SP),                               \
+   jit_ldi_p(FP, 0),                                                    \
+   mz_patch_branch(fp_tmpr),                                            \
+   jit_subi_l(FP, FP, (JIT_FRAME_FLOSTACK_OFFSET - (i))))
+#else
+# define check_fp_depth(i, FP) (void)0
+#endif
+
+#define FLOSTACK_SPACE_CHUNK 16
+# define mz_ld_fppush_x(r, i, FP, extfl) (check_fp_depth(i, FP), jit_FPSEL_ldxi_xd_fppush(extfl, r, FP, (JIT_FRAME_FLOSTACK_OFFSET - (i))))
+# define mz_ld_fppush(r, i, extfl) mz_ld_fppush_x(r, i, JIT_FP, extfl) 
+# define mz_st_fppop_x(i, r, FP, extfl) (check_fp_depth(i, FP), (void)jit_FPSEL_stxi_xd_fppop(extfl, (JIT_FRAME_FLOSTACK_OFFSET - (i)), FP, r))
+# define mz_st_fppop(i, r, extfl) mz_st_fppop_x(i, r, JIT_FP, extfl) 
+
 #define mz_patch_branch(a) mz_patch_branch_at(a, (_jit.x.pc))
 #define mz_patch_ucbranch(a) mz_patch_ucbranch_at(a, (_jit.x.pc))
 
 #ifdef NEED_LONG_JUMPS
 # define __START_SHORT_JUMPS__(cond) if (cond) { _jitl.long_jumps = 0; }
-# define __END_SHORT_JUMPS__(cond) if (cond) { _jitl.long_jumps= 1; }
+# define __END_SHORT_JUMPS__(cond) if (cond) { _jitl.long_jumps = LONG_JUMPS_DEFAULT(_jitl); }
 #else
 # define __START_SHORT_JUMPS__(cond) /* empty */
 # define __END_SHORT_JUMPS__(cond) /* empty */
@@ -867,6 +978,13 @@ void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
 # define mz_bnei_p(a, v, i) jit_bnei_p(a, v, i)
 #endif
 
+#ifdef jit_leai_l
+# define jit_fixnum_l(JIT_Rdest, JIT_Rsrc) jit_leai_l(JIT_Rdest, JIT_Rsrc, 1, 1)
+#else
+# define jit_fixnum_l(JIT_Rdest, JIT_Rsrc) (jit_lshi_l(JIT_Rdest, JIT_Rsrc, 1), \
+                                            jit_ori_l(JIT_Rdest, JIT_Rdest, 0x1))
+#endif
+
 /* 
  About short-jump mode:
    
@@ -885,7 +1003,19 @@ void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
    Tiny-jump mode is like short-jump mode, but the offset must be
    within +/- 2^7. Favor tiny jumps over short jumps when possible.
 
-   All mz_finish() and jit_calli() are implicitly long jumps.
+   On x86_64, short is the default, since "short" is pretty long.
+   Short mode is never needed for jumps within a single allocated
+   block (on the assumption that a single block of code can never get
+   that long). Default-long mode must be enabled if allocated code
+   blocks can be far apart.
+
+   A jit_calli() is "medium": for x86_64, it is short unless
+   default-long mode is enabled; otherwise, it is always
+   long.
+
+   All mz_finish() are long jumps. This is true even in default-short
+   jump mode on x86_64, since the target is likely to be C code that
+   is not necessarily close to JIT-allocate code.
 */
 
 /* A lightweight continuation is one that contains only frames from
@@ -900,7 +1030,7 @@ void scheme_jit_prolog_again(mz_jit_state *jitter, int n, int ret_addr_reg)
      (Other JIT constraints imply that it isn't a pointer to GCable memory.)
 
    * Relevant thread-local state is confined to the C stack, runstack,
-     mark stack, and tl_save_fp.
+     mark stack, and tl_save_fp[2].
 
    * A pointer to the runstack can be used as a Scheme_Object** argument, but
      only when it points to MZ_RUNSTACK.
@@ -1000,6 +1130,11 @@ static void emit_indentation(mz_jit_state *jitter)
 #define JIT_FPR_1(r) JIT_FPR1
 #endif
 
+#ifdef MZ_LONG_DOUBLE
+#define JIT_FPU_FPR_0(r) JIT_FPU_FPR0
+#define JIT_FPU_FPR_1(r) JIT_FPU_FPR1
+#endif
+
 #if defined(MZ_USE_JIT_I386)
 # define mz_movi_d_fppush(rd,immd,tmp)    { GC_CAN_IGNORE void *addr; \
                                             addr = scheme_mz_retain_double(jitter, immd); \
@@ -1009,6 +1144,25 @@ static void emit_indentation(mz_jit_state *jitter)
 # define mz_movi_d_fppush(rd,immd,tmp)    jit_movi_d_fppush(rd,immd)    
 #endif
 
+#ifdef MZ_LONG_DOUBLE
+# define mz_fpu_movi_ld_fppush(rd,immd,tmp)    { GC_CAN_IGNORE void *addr; \
+                                                 addr = scheme_mz_retain_long_double(jitter, immd); \
+                                                 (void)jit_patchable_movi_p(tmp, addr);        \
+                                                 jit_fpu_ldr_ld_fppush(rd, tmp); }
+#endif
+
+
+/**********************************************************************/
+
+/* Does boxing a type require registers, possibly GC, etc.? */
+#ifdef MZ_LONG_DOUBLE
+#define JIT_TYPE_NEEDS_BOXING(t) ((t) == SCHEME_LOCAL_TYPE_FLONUM \
+                                  || (t) == SCHEME_LOCAL_TYPE_EXTFLONUM)
+                                  
+#else
+#define JIT_TYPE_NEEDS_BOXING(t) ((t) == SCHEME_LOCAL_TYPE_FLONUM)
+#endif
+
 /**********************************************************************/
 
 #ifdef MZ_USE_FUTURES
@@ -1073,22 +1227,22 @@ static void emit_indentation(mz_jit_state *jitter)
 /*                             jitstate                               */
 /**********************************************************************/
 
-#ifdef SIXTY_FOUR_BIT_INTEGERS
+#if defined(SIXTY_FOUR_BIT_INTEGERS) || defined(MZ_USE_JIT_PPC)
 # define JIT_BUFFER_PAD_SIZE 200
 #else
 # define JIT_BUFFER_PAD_SIZE 100
 #endif
 
 #define PAST_LIMIT() ((uintptr_t)jit_get_ip().ptr > (uintptr_t)jitter->limit)
-#define CHECK_LIMIT() if (PAST_LIMIT()) return past_limit(jitter);
+#define CHECK_LIMIT() if (PAST_LIMIT()) return past_limit(jitter, __FILE__, __LINE__);
 #if 1
-# define past_limit(j) 0
+# define past_limit(j, f, l) 0
 #else
-static int past_limit(mz_jit_state *jitter)
+static int past_limit(mz_jit_state *jitter, const char *file, int line)
 {
   if (((uintptr_t)jit_get_ip().ptr > (uintptr_t)jitter->limit + JIT_BUFFER_PAD_SIZE)
       || (jitter->retain_start)) {
-    printf("way past\n"); abort();
+    printf("way past %s %d\n", file, line); abort();
   }
   return 0;
 }
@@ -1108,13 +1262,16 @@ int scheme_mz_flostack_save(mz_jit_state *jitter, int *pos);
 int scheme_mz_compute_runstack_restored(mz_jit_state *jitter, int adj, int skip);
 int scheme_mz_retain_it(mz_jit_state *jitter, void *v);
 double *scheme_mz_retain_double(mz_jit_state *jitter, double d);
+#ifdef MZ_LONG_DOUBLE
+long_double *scheme_mz_retain_long_double(mz_jit_state *jitter, long_double d);
+#endif
 int scheme_mz_remap_it(mz_jit_state *jitter, int i);
 void scheme_mz_pushr_p_it(mz_jit_state *jitter, int reg);
 void scheme_mz_popr_p_it(mz_jit_state *jitter, int reg, int discard);
 void scheme_mz_need_space(mz_jit_state *jitter, int need_extra);
 int scheme_stack_safety(mz_jit_state *jitter, int cnt, int offset);
 #ifdef USE_FLONUM_UNBOXING
-int scheme_mz_flonum_pos(mz_jit_state *jitter, int i);
+int scheme_mz_flostack_pos(mz_jit_state *jitter, int i);
 #endif
 void scheme_mz_load_retained(mz_jit_state *jitter, int rs, void *o);
 
@@ -1134,30 +1291,49 @@ int scheme_mz_try_runstack_pop(mz_jit_state *jitter, int n);
 #define mz_runstack_popped(j, n) scheme_mz_runstack_popped(j, n) 
 #define mz_try_runstack_pop(j, n) scheme_mz_try_runstack_pop(j, n) 
 
+typedef struct {
+  int unbox;
+#ifdef MZ_LONG_DOUBLE
+  int unbox_extflonum;
+#endif
+} mz_jit_unbox_state;
+
+void scheme_mz_unbox_save(mz_jit_state *jitter, mz_jit_unbox_state *r);
+void scheme_mz_unbox_restore(mz_jit_state *jitter, mz_jit_unbox_state *r);
+
 /**********************************************************************/
 /*                             jitinline                              */
 /**********************************************************************/
 
 int scheme_inlined_unary_prim(Scheme_Object *o, Scheme_Object *_app, mz_jit_state *jitter);
 int scheme_inlined_binary_prim(Scheme_Object *o, Scheme_Object *_app, mz_jit_state *jitter);
-int scheme_inlined_nary_prim(Scheme_Object *o, Scheme_Object *_app);
+int scheme_inlined_nary_prim(Scheme_Object *o, Scheme_Object *_app, mz_jit_state *jitter);
 int scheme_generate_inlined_unary(mz_jit_state *jitter, Scheme_App2_Rec *app, int is_tail, int multi_ok, 
-				  Branch_Info *for_branch, int branch_short, int need_sync, int result_ignored);
+				  Branch_Info *for_branch, int branch_short, int need_sync, int result_ignored,
+                                  int dest);
 int scheme_generate_inlined_binary(mz_jit_state *jitter, Scheme_App3_Rec *app, int is_tail, int multi_ok, 
-				   Branch_Info *for_branch, int branch_short, int need_sync, int result_ignored);
+				   Branch_Info *for_branch, int branch_short, int need_sync, int result_ignored,
+                                   int dest);
 int scheme_generate_inlined_nary(mz_jit_state *jitter, Scheme_App_Rec *app, int is_tail, int multi_ok, 
-                                 Branch_Info *for_branch, int branch_short, int result_ignored);
+                                 Branch_Info *for_branch, int branch_short, int result_ignored,
+                                 int dest);
 int scheme_generate_inlined_test(mz_jit_state *jitter, Scheme_Object *obj, int branch_short, 
                                  Branch_Info *for_branch, int need_sync);
-int scheme_generate_cons_alloc(mz_jit_state *jitter, int rev, int inline_retry);
+int scheme_generate_cons_alloc(mz_jit_state *jitter, int rev, int inline_retry, int known_list, int dest);
+int scheme_generate_struct_alloc(mz_jit_state *jitter, int num_args, 
+                                 int inline_slow, int pop_and_jump,
+                                 int is_tail, int multi_ok, int dest);
+int scheme_generate_two_args(Scheme_Object *rand1, Scheme_Object *rand2, mz_jit_state *jitter, 
+                             int order_matters, int skipped);
 
 /**********************************************************************/
 /*                             jitalloc                               */
 /**********************************************************************/
 
 #ifdef CAN_INLINE_ALLOC
-int scheme_inline_alloc(mz_jit_state *jitter, int amt, Scheme_Type ty, int immut,
-			int keep_r0_r1, int keep_fpr1, int inline_retry);
+int scheme_inline_alloc(mz_jit_state *jitter, int amt, Scheme_Type ty, int flags,
+			int keep_r0_r1, int keep_fpr1, int inline_retry
+                        , int keep_extfpr1);
 int scheme_generate_alloc_retry(mz_jit_state *jitter, int i);
 #else
 Scheme_Object *scheme_jit_make_list(GC_CAN_IGNORE Scheme_Object **rs, intptr_t n);
@@ -1174,17 +1350,30 @@ Scheme_Object *scheme_jit_make_two_element_ivector(Scheme_Object *a, Scheme_Obje
 /*                             jitarith                               */
 /**********************************************************************/
 
-int scheme_can_unbox_inline(Scheme_Object *obj, int fuel, int regs, int unsafely);
-int scheme_can_unbox_directly(Scheme_Object *obj);
+int scheme_jit_is_fixnum(Scheme_Object *rand);
+int scheme_can_unbox_inline(Scheme_Object *obj, int fuel, int regs, int unsafely, int extfl);
+int scheme_can_unbox_directly(Scheme_Object *obj, int extfl);
 int scheme_generate_unboxing(mz_jit_state *jitter, int target);
 int scheme_generate_pop_unboxed(mz_jit_state *jitter);
 int scheme_generate_nary_arith(mz_jit_state *jitter, Scheme_App_Rec *app,
-                               int arith, int cmp, Branch_Info *for_branch, int branch_short);
-int scheme_generate_alloc_double(mz_jit_state *jitter, int inline_retry);
+                               int arith, int cmp, Branch_Info *for_branch, int branch_short,
+                               int dest);
+int scheme_generate_alloc_double(mz_jit_state *jitter, int inline_retry, int dest);
 int scheme_generate_arith(mz_jit_state *jitter, Scheme_Object *rator, Scheme_Object *rand, Scheme_Object *rand2, 
 			  int orig_args, int arith, int cmp, int v, 
                           Branch_Info *for_branch, int branch_short,
-                          int unsafe_fx, int unsafe_fl, GC_CAN_IGNORE jit_insn *overflow_refslow);
+                          int unsafe_fx, int unsafe_fl, GC_CAN_IGNORE jit_insn *overflow_refslow,
+                          int dest);
+
+#ifdef MZ_LONG_DOUBLE
+int scheme_generate_alloc_long_double(mz_jit_state *jitter, int inline_retry, int dest);
+int scheme_generate_extflonum_arith(mz_jit_state *jitter, Scheme_Object *rator, Scheme_Object *rand, Scheme_Object *rand2, 
+                                      int orig_args, int arith, int cmp, int v, 
+                                      Branch_Info *for_branch, int branch_short, int unsafe_fx, int unsafe_extfl,
+                                      GC_CAN_IGNORE jit_insn *overflow_refslow, int dest);
+#endif
+
+int scheme_generate_alloc_X_double(mz_jit_state *jitter, int inline_retry, int dest, int extfl);
 
 /**********************************************************************/
 /*                              jitcall                               */
@@ -1192,16 +1381,21 @@ int scheme_generate_arith(mz_jit_state *jitter, Scheme_Object *rator, Scheme_Obj
 
 typedef struct jit_direct_arg jit_direct_arg;
 
-void *scheme_generate_shared_call(int num_rands, mz_jit_state *old_jitter, int multi_ok, int is_tail, 
-				  int direct_prim, int direct_native, int nontail_self);
-void scheme_ensure_retry_available(mz_jit_state *jitter, int multi_ok);
+void *scheme_generate_shared_call(int num_rands, mz_jit_state *old_jitter, int multi_ok, int result_ignored, 
+                                  int is_tail, int direct_prim, int direct_native, int nontail_self, int unboxed_args);
+void scheme_ensure_retry_available(mz_jit_state *jitter, int multi_ok, int result_ignored);
 int scheme_generate_app(Scheme_App_Rec *app, Scheme_Object **alt_rands, int num_rands, 
-			mz_jit_state *jitter, int is_tail, int multi_ok, int no_call);
+			mz_jit_state *jitter, int is_tail, int multi_ok, int ignored_result,
+                        int no_call);
 int scheme_generate_tail_call(mz_jit_state *jitter, int num_rands, int direct_native, int need_set_rs, 
                               int is_inline, Scheme_Native_Closure *direct_to_code, jit_direct_arg *direct_arg);
 int scheme_generate_non_tail_call(mz_jit_state *jitter, int num_rands, int direct_native, int need_set_rs, 
-				  int multi_ok, int nontail_self, int pop_and_jump, int is_inlined);
+				  int multi_ok, int result_ignored, int nontail_self, int pop_and_jump, 
+                                  int is_inlined, int unboxed_args);
 int scheme_generate_finish_tail_call(mz_jit_state *jitter, int direct_native);
+int scheme_generate_finish_apply(mz_jit_state *jitter);
+int scheme_generate_finish_multi_apply(mz_jit_state *jitter);
+int scheme_generate_finish_tail_apply(mz_jit_state *jitter);
 void scheme_jit_register_sub_func(mz_jit_state *jitter, void *code, Scheme_Object *protocol);
 void scheme_jit_register_helper_func(mz_jit_state *jitter, void *code);
 #ifdef MZ_USE_FUTURES
@@ -1227,6 +1421,17 @@ void scheme_jit_release_native_code(void *fnlized, void *p);
 int scheme_do_generate_common(mz_jit_state *jitter, void *_data);
 int scheme_do_generate_more_common(mz_jit_state *jitter, void *_data);
 
+int scheme_save_struct_temp(mz_jit_state *jitter, int reg);
+int scheme_restore_struct_temp(mz_jit_state *jitter, int reg);
+int scheme_generate_struct_op(mz_jit_state *jitter, int kind, int for_branch,
+                              Branch_Info *branch_info, int branch_short,
+                              int result_ignored,
+                              int check_proc, int check_arg_fixnum,
+                              int type_pos, int field_pos, 
+                              int pop_and_jump,
+                              jit_insn *refslow, jit_insn *refslow2,
+                              jit_insn *bref_false, jit_insn *bref_true);
+
 /**********************************************************************/
 /*                               jit                                  */
 /**********************************************************************/
@@ -1237,19 +1442,20 @@ int scheme_generate_non_tail_with_branch(Scheme_Object *obj, mz_jit_state *jitte
 int scheme_generate(Scheme_Object *obj, mz_jit_state *jitter, int tail_ok, int wcm_may_replace, int multi_ok, int target,
                     Branch_Info *for_branch);
 int scheme_generate_unboxed(Scheme_Object *obj, mz_jit_state *jitter, int inlined_ok, int unbox_anyway);
-void *scheme_generate_shared_call(int num_rands, mz_jit_state *old_jitter, int multi_ok, int is_tail, 
-				  int direct_prim, int direct_native, int nontail_self);
 
 #ifdef USE_FLONUM_UNBOXING
-int scheme_generate_flonum_local_unboxing(mz_jit_state *jitter, int push);
+int scheme_generate_flonum_local_unboxing(mz_jit_state *jitter, int push, int no_store, int extfl);
+int scheme_generate_flonum_local_boxing(mz_jit_state *jitter, int pos, int offset, int target, int extfl);
 #endif
-int scheme_generate_unboxed(Scheme_Object *obj, mz_jit_state *jitter, int inlined_ok, int unbox_anyway);
 int scheme_generate_non_tail_mark_pos_prefix(mz_jit_state *jitter);
 void scheme_generate_non_tail_mark_pos_suffix(mz_jit_state *jitter);
 
 Scheme_Object **scheme_on_demand(Scheme_Object **argv);
 Scheme_Object **scheme_on_demand_with_args(Scheme_Object **in_argv, Scheme_Object **argv, int argv_delta);
 
+void scheme_jit_allocate_values(int count, Scheme_Thread *p);
+Scheme_Structure *scheme_jit_allocate_structure(int argc, Scheme_Struct_Type *stype);
+
 void scheme_prepare_branch_jump(mz_jit_state *jitter, Branch_Info *for_branch);
 void scheme_branch_for_true(mz_jit_state *jitter, Branch_Info *for_branch);
 void scheme_add_or_patch_branch_true_uc(mz_jit_state *jitter, Branch_Info *for_branch, jit_insn *ref);
@@ -1263,7 +1469,7 @@ int scheme_can_delay_and_avoids_r1(Scheme_Object *obj);
 int scheme_can_delay_and_avoids_r1_r2(Scheme_Object *obj);
 int scheme_is_constant_and_avoids_r1(Scheme_Object *obj);
 int scheme_is_relatively_constant_and_avoids_r1_maybe_fp(Scheme_Object *obj, Scheme_Object *wrt,
-                                                         int fp_ok);
+                                                         int fp_ok, int extfl);
 int scheme_is_relatively_constant_and_avoids_r1(Scheme_Object *obj, Scheme_Object *wrt);
 int scheme_needs_only_target_register(Scheme_Object *obj, int and_can_reorder);
 int scheme_is_noncm(Scheme_Object *a, mz_jit_state *jitter, int depth, int stack_start);
@@ -1275,9 +1481,12 @@ int scheme_is_non_gc(Scheme_Object *obj, int depth);
 int scheme_jit_check_closure_flonum_bit(Scheme_Closure_Data *data, int pos, int delta);
 # define CLOSURE_ARGUMENT_IS_FLONUM(data, pos) scheme_jit_check_closure_flonum_bit(data, pos, 0)
 # define CLOSURE_CONTENT_IS_FLONUM(data, pos) scheme_jit_check_closure_flonum_bit(data, pos, data->num_params)
+ int scheme_jit_check_closure_extflonum_bit(Scheme_Closure_Data *data, int pos, int delta);
+# define CLOSURE_ARGUMENT_IS_EXTFLONUM(data, pos) scheme_jit_check_closure_extflonum_bit(data, pos, 0)
+# define CLOSURE_CONTENT_IS_EXTFLONUM(data, pos) scheme_jit_check_closure_extflonum_bit(data, pos, data->num_params)
 #endif
 
-Scheme_Object *scheme_extract_global(Scheme_Object *o, Scheme_Native_Closure *nc);
+Scheme_Object *scheme_extract_global(Scheme_Object *o, Scheme_Native_Closure *nc, int local_only);
 Scheme_Object *scheme_extract_closure_local(Scheme_Object *obj, mz_jit_state *jitter, int extra_push);
 
 void scheme_jit_register_traversers(void);
@@ -1331,7 +1540,8 @@ Scheme_Object *scheme_jit_continuation_apply_install(Apply_LWC_Args *args);
 #define ARITH_FLUNOP   14
 /*  inexact->exact, unsafe-fl->fx, fl->exact-integer, fl->fx */
 #define ARITH_INEX_EX  15
-
+/*  flexpt */
+#define ARITH_EXPT     16
 
 /* Comparison codes. Used in jitarith.c and jitinline.c. */
 
@@ -1351,3 +1561,17 @@ Scheme_Object *scheme_jit_continuation_apply_install(Apply_LWC_Args *args);
 #define CMP_EVENP  4
 /*  odd? */
 #define CMP_ODDP  -4
+
+/**********************************************************************/
+
+#define INLINE_STRUCT_PROC_PRED 1
+#define INLINE_STRUCT_PROC_GET  2
+#define INLINE_STRUCT_PROC_SET  3
+#define INLINE_STRUCT_PROC_PROP_GET 4
+#define INLINE_STRUCT_PROC_PROP_GET_W_DEFAULT 5
+#define INLINE_STRUCT_PROC_PROP_PRED 6
+#define INLINE_STRUCT_PROC_CONSTR 7
+
+
+
+
