@@ -35,6 +35,9 @@ define_ts_iSi_s(scheme_build_list_offset, FSRC_OTHER)
 #ifdef JITARITH_TS_PROCS
 # if defined(INLINE_FP_OPS) && !defined(CAN_INLINE_ALLOC)
 define_ts__s(malloc_double, FSRC_OTHER)
+# ifdef MZ_LONG_DOUBLE
+   define_ts__s(malloc_long_double, FSRC_OTHER)
+# endif
 # endif
 #endif
 
@@ -74,11 +77,18 @@ define_ts_iS_s(scheme_checked_byte_string_ref, FSRC_MARKS)
 define_ts_iS_s(scheme_checked_byte_string_set, FSRC_MARKS)
 define_ts_iS_s(scheme_checked_flvector_ref, FSRC_MARKS)
 define_ts_iS_s(scheme_checked_flvector_set, FSRC_MARKS)
+#ifdef MZ_LONG_DOUBLE
+define_ts_iS_s(scheme_checked_extflvector_ref, FSRC_MARKS)
+define_ts_iS_s(scheme_checked_extflvector_set, FSRC_MARKS)
+#endif
 define_ts_iS_s(scheme_checked_fxvector_ref, FSRC_MARKS)
 define_ts_iS_s(scheme_checked_fxvector_set, FSRC_MARKS)
 define_ts_iS_s(scheme_checked_syntax_e, FSRC_MARKS)
 define_ts_s_s(scheme_vector_length, FSRC_MARKS)
 define_ts_s_s(scheme_flvector_length, FSRC_MARKS)
+#ifdef MZ_LONG_DOUBLE
+define_ts_s_s(scheme_extflvector_length, FSRC_MARKS)
+#endif
 define_ts_s_s(scheme_fxvector_length, FSRC_MARKS)
 define_ts_s_s(scheme_unbox, FSRC_MARKS)
 define_ts_si_s(scheme_struct_ref, FSRC_MARKS)
@@ -88,10 +98,15 @@ define_ts_iS_s(scheme_procedure_arity_includes, FSRC_MARKS)
 define_ts_ssi_s(vector_check_chaperone_of, FSRC_MARKS)
 define_ts_iS_s(scheme_checked_list_ref, FSRC_MARKS)
 define_ts_iS_s(scheme_checked_list_tail, FSRC_MARKS)
+define_ts_iSs_s(scheme_struct_getter, FSRC_MARKS)
+define_ts_iSs_s(scheme_struct_setter, FSRC_MARKS)
+define_ts_iS_s(scheme_box_cas, FSRC_MARKS)
+define_ts__v(chaperone_set_mark, FSRC_MARKS)
+define_ts_iS_s(scheme_checked_char_to_integer, FSRC_MARKS)
+define_ts_iS_s(scheme_checked_integer_to_char, FSRC_MARKS)
 #endif
 
 #ifdef JITCALL_TS_PROCS
-define_ts_siS_s(_scheme_tail_apply_from_native, FSRC_RATOR)
 define_ts_s_s(scheme_force_value_same_mark, FSRC_MARKS)
 define_ts_s_s(scheme_force_one_value_same_mark, FSRC_MARKS)
 #endif
@@ -111,7 +126,7 @@ define_ts_l_s(scheme_jit_make_ivector, FSRC_OTHER)
 define_ts_l_s(scheme_jit_make_vector, FSRC_OTHER)
 # endif
 define_ts_ss_i(scheme_equal, FSRC_MARKS)
-define_ts_sss_s(scheme_extract_one_cc_mark_to_tag, FSRC_MARKS)
+define_ts_sss_s(extract_one_cc_mark_to_tag, FSRC_MARKS)
 #endif
 
 #ifdef JIT_APPLY_TS_PROCS
@@ -126,13 +141,14 @@ define_ts_s_s(scheme_box, FSRC_OTHER)
 #else
 # define ts__scheme_apply_multi_from_native _scheme_apply_multi_from_native
 # define ts__scheme_apply_from_native _scheme_apply_from_native
-# define ts__scheme_tail_apply_from_native _scheme_tail_apply_from_native
-# define ts__scheme_tail_apply_from_native_fixup_args _scheme_tail_apply_from_native_fixup_args
 # define ts_scheme_force_value_same_mark scheme_force_value_same_mark
 # define ts_scheme_force_one_value_same_mark scheme_force_one_value_same_mark
 # define ts_scheme_force_value_same_mark scheme_force_value_same_mark
 # define ts_scheme_force_one_value_same_mark scheme_force_one_value_same_mark
 # define ts_malloc_double malloc_double
+#ifdef MZ_LONG_DOUBLE
+# define ts_malloc_long_double malloc_long_double
+#endif
 # define ts_scheme_box scheme_box
 # define ts_scheme_make_mutable_pair scheme_make_mutable_pair
 # define ts_scheme_jit_make_list_star scheme_jit_make_list_star
@@ -177,13 +193,18 @@ define_ts_s_s(scheme_box, FSRC_OTHER)
 # define ts_scheme_make_complex scheme_make_complex
 # define ts_scheme_unbox scheme_unbox
 # define ts_scheme_set_box scheme_set_box
+# define ts_scheme_box_cas scheme_box_cas
+# define ts_chaperone_set_mark chaperone_set_mark
 # define ts_scheme_vector_length scheme_vector_length
 # define ts_scheme_flvector_length scheme_flvector_length
+#ifdef MZ_LONG_DOUBLE
+# define ts_scheme_extflvector_length scheme_extflvector_length
+#endif
 # define ts_scheme_fxvector_length scheme_fxvector_length
 # define ts_scheme_struct_ref scheme_struct_ref
 # define ts_scheme_struct_set scheme_struct_set
 # define ts_scheme_equal scheme_equal
-# define ts_scheme_extract_one_cc_mark_to_tag scheme_extract_one_cc_mark_to_tag
+# define ts_extract_one_cc_mark_to_tag extract_one_cc_mark_to_tag
 # define ts_tail_call_with_values_from_multiple_result tail_call_with_values_from_multiple_result
 # define ts_raise_bad_call_with_values raise_bad_call_with_values
 # define ts_call_with_values_from_multiple_result_multi call_with_values_from_multiple_result_multi
@@ -196,6 +217,10 @@ define_ts_s_s(scheme_box, FSRC_OTHER)
 # define ts_scheme_checked_byte_string_set scheme_checked_byte_string_set
 # define ts_scheme_checked_flvector_ref scheme_checked_flvector_ref
 # define ts_scheme_checked_flvector_set scheme_checked_flvector_set
+#ifdef MZ_LONG_DOUBLE
+# define ts_scheme_checked_extflvector_ref scheme_checked_extflvector_ref
+# define ts_scheme_checked_extflvector_set scheme_checked_extflvector_set
+#endif
 # define ts_scheme_checked_fxvector_ref scheme_checked_fxvector_ref
 # define ts_scheme_checked_fxvector_set scheme_checked_fxvector_set
 # define ts_scheme_checked_syntax_e scheme_checked_syntax_e
@@ -207,4 +232,8 @@ define_ts_s_s(scheme_box, FSRC_OTHER)
 # define ts_vector_check_chaperone_of vector_check_chaperone_of
 # define ts_scheme_checked_list_ref scheme_checked_list_ref
 # define ts_scheme_checked_list_tail scheme_checked_list_tail
+# define ts_scheme_struct_getter scheme_struct_getter
+# define ts_scheme_struct_setter scheme_struct_setter
+# define ts_scheme_checked_char_to_integer scheme_checked_char_to_integer
+# define ts_scheme_checked_integer_to_char scheme_checked_integer_to_char
 #endif
