@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2006-2012 PLT Scheme Inc.
+  Copyright (c) 2006-2013 PLT Design Inc.
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Library General Public
@@ -35,6 +35,7 @@
 #ifdef CAN_INLINE_ALLOC
 THREAD_LOCAL_DECL(extern uintptr_t GC_gen0_alloc_page_ptr);
 intptr_t GC_initial_word(int sizeb);
+intptr_t GC_pair_initial_word(int sizeb);
 intptr_t GC_array_initial_word(int sizeb);
 intptr_t GC_compute_alloc_size(intptr_t sizeb);
 
@@ -43,6 +44,10 @@ THREAD_LOCAL_DECL(static void *retry_alloc_r1); /* set by prepare_retry_alloc()
 #ifdef JIT_USE_FP_OPS
 THREAD_LOCAL_DECL(double scheme_jit_save_fp);
 THREAD_LOCAL_DECL(double scheme_jit_save_fp2);
+# ifdef MZ_LONG_DOUBLE
+THREAD_LOCAL_DECL(long_double scheme_jit_save_extfp);
+THREAD_LOCAL_DECL(long_double scheme_jit_save_extfp2);
+# endif
 #endif
 
 static void *prepare_retry_alloc(void *p, void *p2)
@@ -99,17 +104,23 @@ static intptr_t read_first_word(void *sp)
   return foo;
 }
 
-static intptr_t initial_tag_word(Scheme_Type tag, int immut)
+static intptr_t initial_tag_word(Scheme_Type tag, int flags)
 {
-  GC_CAN_IGNORE Scheme_Small_Object sp;
-  memset(&sp, 0, sizeof(Scheme_Small_Object));
+  GC_CAN_IGNORE Scheme_Simple_Object sp;
+  memset(&sp, 0, sizeof(Scheme_Simple_Object));
   sp.iso.so.type = tag;
-  if (immut) SCHEME_SET_IMMUTABLE(&sp);
+  if (flags) {
+    if (tag == scheme_pair_type)
+      SCHEME_PAIR_FLAGS(&sp) |= flags;
+    else
+      SCHEME_SET_IMMUTABLE(&sp);
+  }
   return read_first_word((void *)&sp);
 }
 
-int scheme_inline_alloc(mz_jit_state *jitter, int amt, Scheme_Type ty, int immut,
-			int keep_r0_r1, int keep_fpr1, int inline_retry)
+int scheme_inline_alloc(mz_jit_state *jitter, int amt, Scheme_Type ty, int flags,
+			int keep_r0_r1, int keep_fpr1, int inline_retry,
+                        int keep_extfpr1)
 /* Puts allocated result at JIT_V1; first word is GC tag.
    Uses JIT_R2 as temporary. The allocated memory is "dirty" (i.e., not 0ed).
    Save FP0 when FP ops are enabled. */
@@ -139,6 +150,10 @@ int scheme_inline_alloc(mz_jit_state *jitter, int amt, Scheme_Type ty, int immut
     }
   } else if (keep_fpr1) {
     (void)jit_calli(sjc.retry_alloc_code_keep_fpr1);
+#ifdef MZ_LONG_DOUBLE
+  } else if (keep_extfpr1) {
+    (void)jit_calli(sjc.retry_alloc_code_keep_extfpr1);
+#endif
   } else {
     (void)jit_calli(sjc.retry_alloc_code);
   }
@@ -153,19 +168,21 @@ int scheme_inline_alloc(mz_jit_state *jitter, int amt, Scheme_Type ty, int immut
 
   /* GC header: */
   if (ty >= 0) {
-    a_word = GC_initial_word(amt);
-    jit_movi_l(JIT_R2, a_word);
-    jit_str_l(JIT_V1, JIT_R2);
+    if ((ty == scheme_pair_type)
+        || (ty == scheme_mutable_pair_type)
+        || (ty == scheme_raw_pair_type))
+      a_word = GC_pair_initial_word(amt);
+    else
+      a_word = GC_initial_word(amt);
+    jit_stir_l(JIT_V1, a_word);
     
     /* Scheme_Object header: */
-    a_word = initial_tag_word(ty, immut);
-    jit_movi_l(JIT_R2, a_word);
-    jit_stxi_l(sizeof(intptr_t), JIT_V1, JIT_R2);
+    a_word = initial_tag_word(ty, flags);
+    jit_stixi_l(sizeof(intptr_t), JIT_V1, a_word);
   } else {
     /* an array of pointers */
     a_word = GC_array_initial_word(amt);
-    jit_movi_l(JIT_R2, a_word);
-    jit_str_l(JIT_V1, JIT_R2);
+    jit_stir_l(JIT_V1, a_word);
   }
 
   CHECK_LIMIT();
@@ -198,6 +215,12 @@ static void *malloc_double(void)
 {
   return scheme_make_double(scheme_jit_save_fp);
 }
+#ifdef MZ_LONG_DOUBLE
+static void *malloc_long_double(void)
+{
+  return scheme_make_long_double(scheme_jit_save_extfp);
+}
+#endif
 #endif
 
 #ifdef MZ_PRECISE_GC
@@ -273,6 +296,8 @@ Scheme_Object *scheme_jit_make_two_element_ivector(Scheme_Object *a, Scheme_Obje
 #endif
 
 #ifdef CAN_INLINE_ALLOC
+long_double ld1;
+
 int scheme_generate_alloc_retry(mz_jit_state *jitter, int i)
 {
   GC_CAN_IGNORE jit_insn *refr;
@@ -281,6 +306,11 @@ int scheme_generate_alloc_retry(mz_jit_state *jitter, int i)
   if (i == 2) {
     (void)mz_tl_sti_d_fppop(tl_scheme_jit_save_fp, JIT_FPR0, JIT_R2);
   }
+# ifdef MZ_LONG_DOUBLE
+  if (i == 3) {
+    (void)mz_fpu_tl_sti_ld_fppop(tl_scheme_jit_save_extfp, JIT_FPU_FPR0, JIT_R2);
+  }
+# endif
 #endif
   JIT_UPDATE_THREAD_RSPTR();
   jit_prepare(2);
@@ -302,6 +332,11 @@ int scheme_generate_alloc_retry(mz_jit_state *jitter, int i)
   if (i == 2) {
     (void)mz_tl_ldi_d_fppush(JIT_FPR0, tl_scheme_jit_save_fp, JIT_R2);
   }
+# ifdef MZ_LONG_DOUBLE
+  if (i == 3) {
+    (void)mz_fpu_tl_ldi_ld_fppush(JIT_FPU_FPR0, tl_scheme_jit_save_extfp, JIT_R2);
+  }
+# endif
 #endif
   return 1;
 }
