@@ -20,6 +20,7 @@ module_var {
   gcMARK2(mv->modidx, gc);
   gcMARK2(mv->sym, gc);
   gcMARK2(mv->insp, gc);
+  gcMARK2(mv->shape, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Module_Variable));
@@ -89,6 +90,12 @@ small_object {
   gcBYTES_TO_WORDS(sizeof(Scheme_Small_Object));
 }
 
+small_atomic_obj {
+ mark:
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Small_Object));
+}
+
 app_rec {
   Scheme_App_Rec *r = (Scheme_App_Rec *)p;
 
@@ -467,12 +474,6 @@ escaping_cont_proc {
   gcBYTES_TO_WORDS(sizeof(Scheme_Escaping_Cont));
 }
 
-char_obj {
- mark:
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Small_Object));
-}
-
 bignum_obj {
   Scheme_Bignum *b = (Scheme_Bignum *)p;
 
@@ -516,6 +517,22 @@ double_obj {
   gcBYTES_TO_WORDS(sizeof(Scheme_Double));
 }
 
+#ifdef MZ_LONG_DOUBLE
+long_double_obj {
+ mark:
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Long_Double));
+}
+#else
+long_double_obj {
+ mark:
+  Scheme_Long_Double *ld = (Scheme_Long_Double *)p;
+  gcMARK2(ld->printed_form, gc);
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Long_Double));
+}
+#endif
+
 complex_obj {
  mark:
   Scheme_Complex *c = (Scheme_Complex *)p;
@@ -550,7 +567,7 @@ symbol_obj {
 
  mark:
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Symbol) + s->len - 3);
+  gcBYTES_TO_WORDS(sizeof(Scheme_Symbol) + s->len + 1 - mzFLEX4_DELTA);
 }
 
 cons_cell {
@@ -595,6 +612,17 @@ flvector_obj {
 		    + ((vec->size - mzFLEX_DELTA) * sizeof(double))));
 }
 
+#ifdef MZ_LONG_DOUBLE
+extflvector_obj {
+  Scheme_Long_Double_Vector *vec = (Scheme_Long_Double_Vector *)p;
+
+ mark:
+ size:
+  gcBYTES_TO_WORDS((sizeof(Scheme_Long_Double_Vector) 
+		    + ((vec->size - mzFLEX_DELTA) * sizeof(long double))));
+}
+#endif
+
 input_port {
  mark:
   Scheme_Input_Port *ip = (Scheme_Input_Port *)p;
@@ -949,7 +977,6 @@ prefix_val {
   int i;
   for (i = pf->num_slots; i--; )
     gcMARK2(pf->a[i], gc);
-  gcMARK2(pf->import_map, gc);
  size:
   gcBYTES_TO_WORDS((sizeof(Scheme_Prefix) 
 		    + ((pf->num_slots-mzFLEX_DELTA) * sizeof(Scheme_Object *))
@@ -1018,6 +1045,8 @@ module_val {
  mark:
   Scheme_Module *m = (Scheme_Module *)p;
 
+  gcMARK2(m->phaseless, gc);
+
   gcMARK2(m->code_key, gc);
 
   gcMARK2(m->modname, gc);
@@ -1054,6 +1083,7 @@ module_val {
   gcMARK2(m->submodule_path, gc);
   gcMARK2(m->pre_submodules, gc);
   gcMARK2(m->post_submodules, gc);
+  gcMARK2(m->pre_submodule_names, gc);
   gcMARK2(m->supermodule, gc);
   gcMARK2(m->submodule_ancestry, gc);
 
@@ -1174,6 +1204,7 @@ mark_logger {
   Scheme_Logger *l = (Scheme_Logger *)p;
   gcMARK2(l->name, gc);
   gcMARK2(l->parent, gc);
+  gcMARK2(l->callback, gc);
   gcMARK2(l->timestamp, gc);
   gcMARK2(l->syslog_level, gc);
   gcMARK2(l->stderr_level, gc);
@@ -1310,6 +1341,7 @@ mark_optimize_info {
   gcMARK2(i->transitive_use_len, gc);
   gcMARK2(i->context, gc);
   gcMARK2(i->logger, gc);
+  gcMARK2(i->types, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Optimize_Info));
@@ -1382,7 +1414,7 @@ mark_closure_info {
   
   gcMARK2(i->local_flags, gc);
   gcMARK2(i->base_closure_map, gc);
-  gcMARK2(i->flonum_map, gc);
+  gcMARK2(i->local_type_map, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Closure_Info));
@@ -1451,7 +1483,6 @@ hash_tree_val {
   Scheme_Hash_Tree *ht = (Scheme_Hash_Tree *)p;
 
   gcMARK2(ht->root, gc);
-  gcMARK2(ht->elems_box, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Hash_Tree));
@@ -1484,8 +1515,7 @@ START place;
 place_bi_channel_val {
  mark:
   Scheme_Place_Bi_Channel *pbc = (Scheme_Place_Bi_Channel *)p;
-  gcMARK2(pbc->sendch, gc);
-  gcMARK2(pbc->recvch, gc);
+  gcMARK2(pbc->link, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Place_Bi_Channel));
@@ -2102,17 +2132,6 @@ mark_struct_type_val {
                        * sizeof(Scheme_Struct_Type *))));
 }
 
-mark_struct_proc_info {
- mark:
-  Struct_Proc_Info *i = (Struct_Proc_Info *)p;
-
-  gcMARK2(i->struct_type, gc);
-  gcMARK2(i->func_name, gc);
-
- size:
-  gcBYTES_TO_WORDS(sizeof(Struct_Proc_Info));
-}
-
 mark_struct_property {
  mark:
   Scheme_Struct_Property *i = (Scheme_Struct_Property *)p;
