@@ -1,6 +1,6 @@
 /*
-  Mzscheme
-  Copyright (c) 2004-2012 PLT Scheme Inc.
+  Racket
+  Copyright (c) 2004-2013 PLT Design Inc.
   Copyright (c) 1995 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
