@@ -112,6 +112,7 @@
 #define scheme_log (scheme_extension_table->scheme_log)
 #define scheme_log_w_data (scheme_extension_table->scheme_log_w_data)
 #define scheme_log_message (scheme_extension_table->scheme_log_message)
+#define scheme_log_name_message (scheme_extension_table->scheme_log_name_message)
 #define scheme_log_abort (scheme_extension_table->scheme_log_abort)
 #define scheme_log_warning (scheme_extension_table->scheme_log_warning)
 #define scheme_glib_log_message (scheme_extension_table->scheme_glib_log_message)
@@ -196,6 +197,7 @@
 #define scheme_set_tail_buffer_size (scheme_extension_table->scheme_set_tail_buffer_size)
 #define scheme_force_value (scheme_extension_table->scheme_force_value)
 #define scheme_force_one_value (scheme_extension_table->scheme_force_one_value)
+#define scheme_ignore_result (scheme_extension_table->scheme_ignore_result)
 #define scheme_set_cont_mark (scheme_extension_table->scheme_set_cont_mark)
 #define scheme_push_continuation_frame (scheme_extension_table->scheme_push_continuation_frame)
 #define scheme_pop_continuation_frame (scheme_extension_table->scheme_pop_continuation_frame)
@@ -318,6 +320,7 @@
 #define scheme_prim_is_method (scheme_extension_table->scheme_prim_is_method)
 #define scheme_make_pair (scheme_extension_table->scheme_make_pair)
 #define scheme_make_mutable_pair (scheme_extension_table->scheme_make_mutable_pair)
+#define scheme_make_list_pair (scheme_extension_table->scheme_make_list_pair)
 #define scheme_make_raw_pair (scheme_extension_table->scheme_make_raw_pair)
 #define scheme_make_byte_string (scheme_extension_table->scheme_make_byte_string)
 #define scheme_make_sized_byte_string (scheme_extension_table->scheme_make_sized_byte_string)
@@ -347,6 +350,9 @@
 #define scheme_string_recase (scheme_extension_table->scheme_string_recase)
 #define scheme_make_vector (scheme_extension_table->scheme_make_vector)
 #define scheme_alloc_flvector (scheme_extension_table->scheme_alloc_flvector)
+#ifdef MZ_LONG_DOUBLE
+#define scheme_alloc_extflvector (scheme_extension_table->scheme_alloc_extflvector)
+#endif
 #define scheme_alloc_fxvector (scheme_extension_table->scheme_alloc_fxvector)
 #define scheme_make_integer_value (scheme_extension_table->scheme_make_integer_value)
 #define scheme_make_integer_value_from_unsigned (scheme_extension_table->scheme_make_integer_value_from_unsigned)
@@ -355,6 +361,9 @@
 #define scheme_make_integer_value_from_long_halves (scheme_extension_table->scheme_make_integer_value_from_long_halves)
 #define scheme_make_integer_value_from_unsigned_long_halves (scheme_extension_table->scheme_make_integer_value_from_unsigned_long_halves)
 #define scheme_make_double (scheme_extension_table->scheme_make_double)
+#ifdef MZ_LONG_DOUBLE
+#define scheme_make_long_double (scheme_extension_table->scheme_make_long_double)
+#endif
 #ifdef MZ_USE_SINGLE_FLOATS
 #define scheme_make_float (scheme_extension_table->scheme_make_float)
 #endif
@@ -373,6 +382,9 @@
 #define scheme_get_long_long_val (scheme_extension_table->scheme_get_long_long_val)
 #define scheme_get_unsigned_long_long_val (scheme_extension_table->scheme_get_unsigned_long_long_val)
 #define scheme_real_to_double (scheme_extension_table->scheme_real_to_double)
+#ifdef MZ_LONG_DOUBLE
+#define scheme_real_to_long_double (scheme_extension_table->scheme_real_to_long_double)
+#endif
 #define scheme_make_cptr (scheme_extension_table->scheme_make_cptr)
 #define scheme_make_offset_cptr (scheme_extension_table->scheme_make_offset_cptr)
 #define scheme_make_external_cptr (scheme_extension_table->scheme_make_external_cptr)
@@ -400,6 +412,10 @@
 #define scheme_make_bignum_from_unsigned_long_long (scheme_extension_table->scheme_make_bignum_from_unsigned_long_long)
 #define scheme_bignum_to_double (scheme_extension_table->scheme_bignum_to_double)
 #define scheme_bignum_from_double (scheme_extension_table->scheme_bignum_from_double)
+#ifdef MZ_LONG_DOUBLE
+#define scheme_bignum_to_long_double (scheme_extension_table->scheme_bignum_to_long_double)
+#define scheme_bignum_from_long_double (scheme_extension_table->scheme_bignum_from_long_double)
+#endif
 #ifdef MZ_USE_SINGLE_FLOATS
 #define scheme_bignum_to_float (scheme_extension_table->scheme_bignum_to_float)
 #define scheme_bignum_from_float (scheme_extension_table->scheme_bignum_from_float)
@@ -415,6 +431,10 @@
 #define scheme_make_rational (scheme_extension_table->scheme_make_rational)
 #define scheme_rational_to_double (scheme_extension_table->scheme_rational_to_double)
 #define scheme_rational_from_double (scheme_extension_table->scheme_rational_from_double)
+#ifdef MZ_LONG_DOUBLE
+#define scheme_rational_to_long_double (scheme_extension_table->scheme_rational_to_long_double)
+#define scheme_rational_from_long_double (scheme_extension_table->scheme_rational_from_long_double)
+#endif
 #ifdef MZ_USE_SINGLE_FLOATS
 #define scheme_rational_to_float (scheme_extension_table->scheme_rational_to_float)
 #define scheme_rational_from_float (scheme_extension_table->scheme_rational_from_float)
