@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2012 PLT Scheme Inc.
+  Copyright (c) 2004-2013 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -354,16 +354,16 @@ void scheme_uncopy_stack(int ok, Scheme_Jumpup_Buf *b, intptr_t *prev)
 
     if (c->cont) {
 #ifdef STACK_GROWS_UP
-      top_delta = (((uintptr_t)c->cont->buf.stack_from
+      top_delta = (((uintptr_t)c->cont->buf_ptr->buf.stack_from
 		    + c->cont->buf.stack_size)
 		   - (uintptr_t)c->stack_from);
 #else
       bottom_delta = ((uintptr_t)c->stack_from 
 		      + c->stack_size
-		      - (uintptr_t)c->cont->buf.stack_from);
+		      - (uintptr_t)c->cont->buf_ptr->buf.stack_from);
       top_delta = bottom_delta;
 #endif
-      c = &c->cont->buf;
+      c = &c->cont->buf_ptr->buf;
     } else
       c = NULL;
   }
@@ -518,12 +518,15 @@ int scheme_setjmpup_relative(Scheme_Jumpup_Buf *b, void *base,
          but watch out. */
       intptr_t same_size;
       START_XFORM_SKIP;
-      same_size = find_same(get_copy(c->buf.stack_copy), c->buf.stack_from, c->buf.stack_size);
+      same_size = find_same(get_copy(c->buf_ptr->buf.stack_copy), 
+                            c->buf_ptr->buf.stack_from, 
+                            c->buf_ptr->buf.stack_size);
       b->cont = c;
 #ifdef STACK_GROWS_UP
-      start = (void *)((char *)c->buf.stack_from + same_size);
+      start = (void *)((char *)c->buf_ptr->buf.stack_from + same_size);
 #else
-      start = (void *)((char *)c->buf.stack_from + (c->buf.stack_size - same_size));
+      start = (void *)((char *)c->buf_ptr->buf.stack_from 
+                       + (c->buf_ptr->buf.stack_size - same_size));
 #endif
       /* In 3m-mode, we need `start' on a var-stack boundary: */
       ALIGN_VAR_STACK(__gc_var_stack__, start);
@@ -654,11 +657,23 @@ void scheme_reset_jmpup_buf(Scheme_Jumpup_Buf *b)
 #ifdef USE_MZ_CYGWIN_SETJMP
 /* We have to define setjmp & longjmp to remain compatible
    with MSVC-compiled extensions. It's the mostly same code 
-   as mzsj86.c, just in a slightly different syntax, and it
-   probably only works with -O2. */
+   as mzsj86.c, just in a slightly different syntax. This code
+   is fragile, because it's not well defined whether the compiler
+   will generate frame-pointer setup; use mzsj86g.S, instead. */
+
+#ifdef __MINGW32__
+# if __OPTIMIZE__ > 0
+#  define NEED_STACK_FRAME_SETUP
+# endif
+#endif
 
 int scheme_mz_setjmp(mz_pre_jmp_buf b)
 {
+#ifdef NEED_STACK_FRAME_SETUP
+  asm("push %EBP");
+  asm("mov %ESP, %EBP");
+#endif
+
   asm("mov 4(%EBP), %ECX"); /* return address */
   asm("mov 8(%EBP), %EAX"); /* jmp_buf ptr */
   asm("mov (%EBP), %EDX");  /* old EBP */
@@ -668,12 +683,21 @@ int scheme_mz_setjmp(mz_pre_jmp_buf b)
   asm("mov %ESI, 12(%EAX)");
   asm("mov %ESP, 16(%EAX)");
   asm("mov %ECX, 20(%EAX)");
- 
+
+#ifdef NEED_STACK_FRAME_SETUP
+  asm("pop %EBP");
+#endif
+
   return 0;
 }
 
 void scheme_mz_longjmp(mz_pre_jmp_buf b, int v)
 {
+#ifdef NEED_STACK_FRAME_SETUP
+  asm("push %EBP");
+  asm("mov %ESP, %EBP");
+#endif
+
   asm("mov 12(%EBP), %EAX"); /* return value */
   asm("mov 8(%EBP), %ECX");  /* jmp_buf */
   asm("mov 16(%ECX), %ESP"); /* restore stack pointer */
@@ -685,6 +709,10 @@ void scheme_mz_longjmp(mz_pre_jmp_buf b, int v)
   asm("mov 12(%ECX), %ESI");
   asm("mov 20(%ECX), %ECX"); /* return address */
   asm("mov %ECX, 4(%EBP)");
+
+#ifdef NEED_STACK_FRAME_SETUP
+  asm("pop %EBP");
+#endif
 }
 
 #endif
