@@ -1,8 +1,8 @@
 /*
   Racket
-  Copyright (c) 2004-2011 PLT Scheme Inc.
+  Copyright (c) 2004-2013 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
- 
+
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Library General Public
     License as published by the Free Software Foundation; either
@@ -35,6 +35,8 @@
 static void register_traversers(void);
 #endif
 
+#define FAR_VALUE_FOR_MAX_USED 0x3FFFFFFe
+
 void scheme_init_sfs()
 {
 #ifdef MZ_PRECISE_GC
@@ -172,6 +174,11 @@ void scheme_sfs_used(SFS_Info *info, int pos)
     scheme_signal_error("internal error: misuse of toplevel pointer");
 
   SFS_LOG(printf("touch %d %d\n", pos, info->ip));
+
+  if (info->max_used[pos] >= FAR_VALUE_FOR_MAX_USED) {
+    info->max_used[pos] = (FAR_VALUE_FOR_MAX_USED + 1);
+    return;
+  }
   
   if ((info->min_touch == -1)
       || (pos < info->min_touch))
@@ -209,14 +216,17 @@ Scheme_Object *scheme_sfs_add_clears(Scheme_Object *expr, Scheme_Object *clears,
   return (Scheme_Object *)s;
 }
 
-static void sfs_note_app(SFS_Info *info, Scheme_Object *rator)
+static void sfs_note_app(SFS_Info *info, Scheme_Object *rator, int flags)
 {
   if (!info->pass) {
     if (!info->tail_pos) {
+      if (flags & APPN_FLAG_IMMED)
+        return;
       if (SAME_OBJ(scheme_values_func, rator))
         /* no need to clear for app of `values' */
         return;
       if (SCHEME_PRIMP(rator)) {
+        /* Double-check for immediate primitives: */
         int opt;
         opt = ((Scheme_Prim_Proc_Header *)rator)->flags & SCHEME_PRIM_OPT_MASK;
         if (opt >= SCHEME_PRIM_OPT_IMMEDIATE)
@@ -225,19 +235,22 @@ static void sfs_note_app(SFS_Info *info, Scheme_Object *rator)
       }
       info->max_nontail = info->ip;
     } else {
+      int tail_ok = (flags & APPN_FLAG_SFS_TAIL);      
       if (!MAX_SFS_CLEARING && (info->selfpos >= 0)) {
-        if (SAME_TYPE(SCHEME_TYPE(rator), scheme_local_type)) {
-          if ((SCHEME_LOCAL_POS(rator) + info->stackpos) == info->selfpos) {
-            /* No point in clearing out any of the closure before the
-               tail call. */
-            int i;
-            for (i = info->selflen; i--; ) {
-              if ((info->selfstart + i) != info->tlpos)
-                scheme_sfs_used(info, (info->selfstart - info->stackpos) + i);
-            }
+        if (SAME_TYPE(SCHEME_TYPE(rator), scheme_local_type)
+            && (SCHEME_LOCAL_POS(rator) + info->stackpos) == info->selfpos) {
+          /* No point in clearing out any of the closure before the
+             tail call. */
+          int i;
+          for (i = info->selflen; i--; ) {
+            if ((info->selfstart + i) != info->tlpos)
+              scheme_sfs_used(info, (info->selfstart - info->stackpos) + i);
           }
+          tail_ok = 1;
         }
       }
+      if (!tail_ok)
+        info->max_nontail = info->ip;
     }
   }
 }
@@ -260,7 +273,7 @@ static Scheme_Object *sfs_application(Scheme_Object *o, SFS_Info *info)
     app->args[i] = naya;
   }
 
-  sfs_note_app(info, app->args[0]);
+  sfs_note_app(info, app->args[0], SCHEME_APPN_FLAGS(app) & APPN_FLAG_MASK);
 
   scheme_finish_application(app);
 
@@ -282,7 +295,7 @@ static Scheme_Object *sfs_application2(Scheme_Object *o, SFS_Info *info)
   app->rator = nrator;
   app->rand = nrand;
 
-  sfs_note_app(info, app->rator);
+  sfs_note_app(info, app->rator, SCHEME_APPN_FLAGS(app) & APPN_FLAG_MASK);
 
   scheme_reset_app2_eval_type(app);
   
@@ -307,14 +320,52 @@ static Scheme_Object *sfs_application3(Scheme_Object *o, SFS_Info *info)
   app->rand1 = nrand1;
   app->rand2 = nrand2;
 
-  sfs_note_app(info, app->rator);
+  sfs_note_app(info, app->rator, SCHEME_APPN_FLAGS(app) & APPN_FLAG_MASK);
 
   scheme_reset_app3_eval_type(app);
 
   return o;
 }
 
-static Scheme_Object *sfs_sequence(Scheme_Object *o, SFS_Info *info)
+static Scheme_Object *flatten_sequence(Scheme_Object *o)
+{
+  /* At this point, we sometimes have (begin ... (begin ... (begin ...))).
+     Flatten those out. */
+  Scheme_Sequence *s = (Scheme_Sequence *)o, *s2;
+  int i, extra = 0;
+
+  o = s->array[s->count - 1];
+
+  while (SAME_TYPE(SCHEME_TYPE(o), scheme_sequence_type)) {
+    s2 = (Scheme_Sequence *)o;
+    extra += s2->count - 1;
+    o = s2->array[s2->count - 1];
+  }
+
+  if (extra) {
+    s2 = scheme_malloc_sequence(s->count + extra);
+    s2->so.type = scheme_sequence_type;
+    s2->count = s->count + extra;
+
+    extra = 0;
+    o = (Scheme_Object *)s;
+    while (SAME_TYPE(SCHEME_TYPE(o), scheme_sequence_type)) {
+      s = (Scheme_Sequence *)o;
+      for (i = 0; i < s->count - 1; i++) {
+        s2->array[extra++] = s->array[i];
+      }
+      o = s->array[i];
+    }
+    s2->array[extra++] = o;
+
+    if (extra != s2->count) scheme_signal_error("internal error: flatten failed");
+
+    return (Scheme_Object *)s2;
+  } else
+    return (Scheme_Object *)s;
+}
+
+static Scheme_Object *sfs_sequence(Scheme_Object *o, SFS_Info *info, int can_flatten)
 {
   Scheme_Object *orig, *naya;
   Scheme_Sequence *seq;
@@ -327,10 +378,13 @@ static Scheme_Object *sfs_sequence(Scheme_Object *o, SFS_Info *info)
 
   for (i = 0; i < n; i++) {
     orig = seq->array[i];
-    naya = scheme_sfs_expr(orig, info, -1);
+    naya = scheme_sfs_expr(orig, info, -2);
     seq->array[i] = naya;
   }
 
+  if (can_flatten && info->pass)
+    o = flatten_sequence(o);
+
   return o;
 }
 
@@ -450,7 +504,7 @@ static Scheme_Object *sfs_one_branch(SFS_Info *info, int ip,
         n = info->max_used[i + t_min_t];
         SFS_LOG(printf("%d %s %d %d -> %d/%d\n", info->pass, (delta ? "else" : "then"), ip, 
                        i + t_min_t, n, info->max_calls[i+ t_min_t]));
-        if (n > ip) {
+        if ((n > ip) && (n < FAR_VALUE_FOR_MAX_USED)) {
           SCHEME_VEC_ELS(t_vec)[i] = scheme_make_integer(n);
           info->max_used[i + t_min_t] = ip;
         } else {
@@ -593,6 +647,10 @@ static Scheme_Object *sfs_let_one(Scheme_Object *o, SFS_Info *info)
   save_mnt = info->max_nontail;
 
   if (!info->pass) {
+    if (SCHEME_LET_ONE_TYPE(lo)) {
+      /* never clear a typed slot */
+      info->max_used[pos] = FAR_VALUE_FOR_MAX_USED;
+    }
     vec = scheme_make_vector(3, NULL);
     scheme_sfs_save(info, vec);
   } else {
@@ -611,7 +669,7 @@ static Scheme_Object *sfs_let_one(Scheme_Object *o, SFS_Info *info)
   if (!info->pass)
     info->max_nontail = info->ip;
 # endif
-  
+
   if (!info->pass) {
     int n;
     info->max_calls[pos] = info->max_nontail;
@@ -623,14 +681,15 @@ static Scheme_Object *sfs_let_one(Scheme_Object *o, SFS_Info *info)
   } else {
     info->max_nontail = save_mnt;
 
-    if (info->max_used[pos] <= ip) {
+    if ((info->max_used[pos] <= ip) 
+        || (info->max_used[pos] == FAR_VALUE_FOR_MAX_USED)) {
       /* No one is using it, so don't actually push the value at run time
          (but keep the check that the result is single-valued).
          The optimizer normally would have converted away the binding, but
          it might not because (1) it was introduced late by inlining,
          or (2) the rhs expression doesn't always produce a single
          value. */
-      if (scheme_omittable_expr(rhs, 1, -1, 1, NULL, -1)) {
+      if (scheme_omittable_expr(rhs, 1, -1, 1, NULL, NULL, -1, 0)) {
         rhs = scheme_false;
       } else if ((ip < info->max_calls[pos])
                  && SAME_TYPE(SCHEME_TYPE(rhs), scheme_toplevel_type)) {
@@ -656,7 +715,7 @@ static Scheme_Object *sfs_let_one(Scheme_Object *o, SFS_Info *info)
 
   et = scheme_get_eval_type(lo->value);
   SCHEME_LET_EVAL_TYPE(lo) = (et 
-                              | (unused ? 0 : (SCHEME_LET_EVAL_TYPE(lo) & LET_ONE_FLONUM))
+                              | (unused ? 0 : (SCHEME_LET_EVAL_TYPE(lo) & LET_ONE_TYPE_MASK))
                               | (unused ? LET_ONE_UNUSED : 0));
 
   return o;
@@ -781,6 +840,17 @@ define_values_sfs(Scheme_Object *data, SFS_Info *info)
   return data;
 }
 
+static Scheme_Object *
+inline_variant_sfs(Scheme_Object *data, SFS_Info *info)
+{
+  Scheme_Object *e;
+  scheme_sfs_start_sequence(info, 1, 0);
+  e = scheme_sfs_expr(SCHEME_VEC_ELS(data)[0], info, -1);
+  SCHEME_VEC_ELS(data)[0] = e;
+  /* we don't bother with inlinable variant, since it isn't called directly */
+  return data;
+}
+
 static Scheme_Object *
 set_sfs(Scheme_Object *data, SFS_Info *info)
 {
@@ -1021,6 +1091,25 @@ static Scheme_Object *sfs_closure(Scheme_Object *expr, SFS_Info *info, int self_
       }
     }
 
+    /* Never clear typed arguments or typed closure elements: */
+    if (SCHEME_CLOSURE_DATA_FLAGS(data) & CLOS_HAS_TYPED_ARGS) {
+      int delta, size, ct, j, pos;
+      mzshort *map;
+      delta = data->closure_size;
+      size = data->closure_size + data->num_params;
+      map = data->closure_map;
+      for (j = 0; j < size; j++) {
+        ct = scheme_boxmap_get(map, j, delta);
+        if (ct > CLOS_TYPE_TYPE_OFFSET) {
+          if (j < data->num_params)
+            pos = info->stackpos + delta + j;
+          else
+            pos = info->stackpos + (j - data->num_params);
+          info->max_used[pos] = FAR_VALUE_FOR_MAX_USED;
+        }
+      }
+    }
+
     code = scheme_sfs(data->code, info, data->max_let_depth);
 
     /* If any arguments go unused, and if there's a non-tail,
@@ -1117,6 +1206,7 @@ top_level_require_sfs(Scheme_Object *data, SFS_Info *rslv)
 /*========================================================================*/
 
 Scheme_Object *scheme_sfs_expr(Scheme_Object *expr, SFS_Info *info, int closure_self_pos)
+/* closure_self_pos == -2 => immediately in sequence */
 {
   Scheme_Type type = SCHEME_TYPE(expr);
   int seqn, stackpos, tp;
@@ -1135,7 +1225,7 @@ Scheme_Object *scheme_sfs_expr(Scheme_Object *expr, SFS_Info *info, int closure_
   case scheme_local_unbox_type:
     if (!info->pass)
       scheme_sfs_used(info, SCHEME_LOCAL_POS(expr));
-    else if (SCHEME_GET_LOCAL_FLAGS(expr) != SCHEME_LOCAL_FLONUM) {
+    else if (!SCHEME_GET_LOCAL_TYPE(expr)) {
       int pos, at_ip;
       pos = SCHEME_LOCAL_POS(expr);
       at_ip = info->max_used[info->stackpos + pos];
@@ -1164,8 +1254,10 @@ Scheme_Object *scheme_sfs_expr(Scheme_Object *expr, SFS_Info *info, int closure_
     expr = sfs_application3(expr, info);
     break;
   case scheme_sequence_type:
+    expr = sfs_sequence(expr, info, closure_self_pos != -2);
+    break;
   case scheme_splice_sequence_type:
-    expr = sfs_sequence(expr, info);
+    expr = sfs_sequence(expr, info, 0);
     break;
   case scheme_branch_type:
     expr = sfs_branch(expr, info);
@@ -1250,6 +1342,9 @@ Scheme_Object *scheme_sfs_expr(Scheme_Object *expr, SFS_Info *info, int closure_
   case scheme_module_type:
     expr = module_sfs(expr, info);
     break;
+  case scheme_inline_variant_type:
+    expr = inline_variant_sfs(expr, info);
+    break;
   default:
     break;
   }
