@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2012 PLT Scheme Inc.
+  Copyright (c) 2004-2013 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -161,6 +161,7 @@ scheme_init_type ()
   set_name(scheme_box_type, "<box>");
   set_name(scheme_integer_type, "<fixnum-integer>");
   set_name(scheme_double_type, "<inexact-number>");
+  set_name(scheme_long_double_type, "<extflonum>");
   set_name(scheme_float_type, "<inexact-number*>");
   set_name(scheme_undefined_type, "<undefined>");
   set_name(scheme_eof_type, "<eof>");
@@ -187,6 +188,7 @@ scheme_init_type ()
   set_name(scheme_macro_type, "<macro>");
   set_name(scheme_vector_type, "<vector>");
   set_name(scheme_flvector_type, "<flvector>");
+  set_name(scheme_extflvector_type, "<extflvector>");
   set_name(scheme_fxvector_type, "<fxvector>");
   set_name(scheme_bignum_type, "<bignum-integer>");
   set_name(scheme_escaping_cont_type, "<escape-continuation>");
@@ -295,6 +297,7 @@ scheme_init_type ()
   set_name(scheme_log_reader_type, "<log-receiver>");
 
   set_name(scheme_future_type, "<future>");
+  set_name(scheme_fsemaphore_type, "<fsemaphore>");
 
   set_name(_scheme_values_types_, "<resurrected>");
   set_name(_scheme_compiled_values_types_, "<internal>");
@@ -306,6 +309,8 @@ scheme_init_type ()
 
   set_name(scheme_resolved_module_path_type, "<resolve-module-path>");
 
+  set_name(scheme_phantom_bytes_type, "<phantom-bytes>");
+
 #ifdef MZ_GC_BACKTRACE
   set_name(scheme_rt_meta_cont, "<meta-continuation>");
 #endif
@@ -593,12 +598,13 @@ void scheme_register_traversers(void)
   GC_REG_TRAV(scheme_escaping_cont_type, escaping_cont_proc);
   GC_REG_TRAV(scheme_rt_cont_jmp, cont_jmp_proc);
 
-  GC_REG_TRAV(scheme_char_type, char_obj);
+  GC_REG_TRAV(scheme_char_type, small_atomic_obj);
   GC_REG_TRAV(scheme_integer_type, bad_trav);
   GC_REG_TRAV(scheme_bignum_type, bignum_obj);
   GC_REG_TRAV(scheme_rational_type, rational_obj);
   GC_REG_TRAV(scheme_float_type,  float_obj);
   GC_REG_TRAV(scheme_double_type, double_obj);
+  GC_REG_TRAV(scheme_long_double_type, long_double_obj);
   GC_REG_TRAV(scheme_complex_type, complex_obj);
   GC_REG_TRAV(scheme_char_string_type, string_obj);
   GC_REG_TRAV(scheme_byte_string_type, bstring_obj);
@@ -610,12 +616,15 @@ void scheme_register_traversers(void)
   GC_REG_TRAV(scheme_place_dead_type, small_object);
 #endif
   GC_REG_TRAV(scheme_keyword_type, symbol_obj);
-  GC_REG_TRAV(scheme_null_type, char_obj); /* small */
+  GC_REG_TRAV(scheme_null_type, small_atomic_obj);
   GC_REG_TRAV(scheme_pair_type, cons_cell);
   GC_REG_TRAV(scheme_mutable_pair_type, cons_cell);
   GC_REG_TRAV(scheme_raw_pair_type, cons_cell);
   GC_REG_TRAV(scheme_vector_type, vector_obj);
   GC_REG_TRAV(scheme_flvector_type, flvector_obj);
+#ifdef MZ_LONG_DOUBLE
+  GC_REG_TRAV(scheme_extflvector_type, extflvector_obj);
+#endif
   GC_REG_TRAV(scheme_fxvector_type, fxvector_obj);
   GC_REG_TRAV(scheme_cpointer_type, cpointer_obj);
 
@@ -623,10 +632,10 @@ void scheme_register_traversers(void)
 
   GC_REG_TRAV(scheme_input_port_type, input_port);
   GC_REG_TRAV(scheme_output_port_type, output_port);
-  GC_REG_TRAV(scheme_eof_type, char_obj); /* small */
-  GC_REG_TRAV(scheme_true_type, char_obj); /* small */
-  GC_REG_TRAV(scheme_false_type, char_obj); /* small */
-  GC_REG_TRAV(scheme_void_type, char_obj);  /* small */
+  GC_REG_TRAV(scheme_eof_type, small_atomic_obj);
+  GC_REG_TRAV(scheme_true_type, small_atomic_obj);
+  GC_REG_TRAV(scheme_false_type, small_atomic_obj);
+  GC_REG_TRAV(scheme_void_type, small_atomic_obj); 
   GC_REG_TRAV(scheme_syntax_compiler_type, syntax_compiler);
   GC_REG_TRAV(scheme_macro_type, small_object);
   GC_REG_TRAV(scheme_box_type, small_object);
@@ -653,7 +662,7 @@ void scheme_register_traversers(void)
 
   GC_REG_TRAV(scheme_eval_waiting_type, bad_trav);
   GC_REG_TRAV(scheme_tail_call_waiting_type, bad_trav);
-  GC_REG_TRAV(scheme_undefined_type, char_obj); /* small */
+  GC_REG_TRAV(scheme_undefined_type, small_atomic_obj);
   GC_REG_TRAV(scheme_placeholder_type, small_object);
   GC_REG_TRAV(scheme_table_placeholder_type, iptr_obj);
 
@@ -672,9 +681,9 @@ void scheme_register_traversers(void)
   GC_REG_TRAV(scheme_security_guard_type, guard_val);
 
   GC_REG_TRAV(scheme_nack_evt_type, twoptr_obj);
-  GC_REG_TRAV(scheme_always_evt_type, char_obj);
-  GC_REG_TRAV(scheme_never_evt_type, char_obj);
-  GC_REG_TRAV(scheme_thread_recv_evt_type, char_obj);
+  GC_REG_TRAV(scheme_always_evt_type, small_atomic_obj);
+  GC_REG_TRAV(scheme_never_evt_type, small_atomic_obj);
+  GC_REG_TRAV(scheme_thread_recv_evt_type, small_atomic_obj);
   GC_REG_TRAV(scheme_port_closed_evt_type, small_object);
 
   GC_REG_TRAV(scheme_inspector_type, mark_inspector);
@@ -708,6 +717,9 @@ void scheme_register_traversers(void)
   GC_REG_TRAV(scheme_rib_delimiter_type, small_object);
   GC_REG_TRAV(scheme_noninline_proc_type, small_object);
   GC_REG_TRAV(scheme_prune_context_type, small_object);
+
+  GC_REG_TRAV(scheme_proc_shape_type, small_object);
+  GC_REG_TRAV(scheme_struct_proc_shape_type, small_atomic_obj);
 }
 
 END_XFORM_SKIP;
