@@ -37,8 +37,10 @@
 
 #define USE_GETRUSAGE
 
-#if defined(MZ_USE_PLACES) || defined(USE_PTHREAD_INSTEAD_OF_ITIMER)
-# define USE_PTHREAD_THREAD_TIMER
-#else
-# define USE_ITIMER
+#ifndef MZ_USE_DETERMINSTIC_FUEL
+# if defined(MZ_USE_PLACES) || defined(USE_PTHREAD_INSTEAD_OF_ITIMER)
+#  define USE_PTHREAD_THREAD_TIMER
+# else
+#  define USE_ITIMER
+# endif
 #endif
